@@ -22,8 +22,12 @@
 #include <mousepad/mousepad-replace-dialog.h>
 #include <mousepad/mousepad-window.h>
 #include <mousepad/mousepad-util.h>
+#include <mousepad/mousepad-plugin-provider.h>
+#include <mousepad/mousepad-history.h>
 
-#include <xfconf/xfconf.h>
+
+
+#define DEFAULT_FONT "Monospace 10"
 
 
 
@@ -68,6 +72,9 @@ static void        mousepad_application_set_shared_menu_parts     (MousepadAppli
                                                                    GMenuModel               *model);
 static void        mousepad_application_create_languages_menu     (MousepadApplication      *application);
 static void        mousepad_application_create_style_schemes_menu (MousepadApplication      *application);
+static void        mousepad_application_action_preferences        (GSimpleAction            *action,
+                                                                   GVariant                 *value,
+                                                                   gpointer                  data);
 static void        mousepad_application_action_quit               (GSimpleAction            *action,
                                                                    GVariant                 *value,
                                                                    gpointer                  data);
@@ -77,9 +84,13 @@ static void        mousepad_application_toggle_activate           (GSimpleAction
 static void        mousepad_application_radio_activate            (GSimpleAction            *action,
                                                                    GVariant                 *parameter,
                                                                    gpointer                  data);
+static void        mousepad_application_plugin_activate           (GSimpleAction            *action,
+                                                                   GVariant                 *parameter,
+                                                                   gpointer                  data);
 static void        mousepad_application_action_update             (MousepadApplication      *application,
                                                                    gchar                    *key,
                                                                    GSettings                *settings);
+static void        mousepad_application_plugin_update             (MousepadApplication      *application);
 static void        mousepad_application_action_whitespace         (GSimpleAction            *action,
                                                                    GVariant                 *state,
                                                                    gpointer                  data);
@@ -95,19 +106,27 @@ struct _MousepadApplication
 {
   GtkApplication      __parent__;
 
-  /* preferences dialog related */
-  GtkWidget                   *prefs_dialog;
-  GtkSourceSpaceLocationFlags  space_location_flags;
+  /* preferences dialog */
+  GtkWidget *prefs_dialog;
+  gboolean   prefs_dialog_standalone;
 
   /* command line options */
   gint             opening_mode, line, column;
   MousepadEncoding encoding;
+
+  /* properties */
+  gchar                       *default_font;
+  GtkSourceSpaceLocationFlags  space_location_flags;
+
+  /* plugins */
+  GList *providers;
 };
 
 /* MousepadApplication properties */
 enum
 {
-  PROP_SPACE_LOCATION=1,
+  PROP_DEFAULT_FONT = 1,
+  PROP_SPACE_LOCATION,
   N_PROPERTIES
 };
 
@@ -139,6 +158,11 @@ static const GOptionEntry option_entries[] =
     N_("Column number the cursor to position to (COLUMN >= 0 from start, COLUMN < 0 from end)"),
     N_("COLUMN")
   },
+  {
+    "preferences", '\0', G_OPTION_FLAG_NONE,
+    G_OPTION_ARG_NONE, NULL,
+    N_("Open the preferences dialog"), NULL
+  },
   {
     "disable-server", '\0', G_OPTION_FLAG_NONE,
     G_OPTION_ARG_NONE, NULL,
@@ -178,6 +202,7 @@ enum
 static const GActionEntry stateless_actions[] =
 {
   /* command line options */
+  { "preferences", mousepad_application_action_preferences, NULL, NULL, NULL },
   { "quit", mousepad_application_action_quit, NULL, NULL, NULL }
 };
 #define N_STATELESS G_N_ELEMENTS (stateless_actions)
@@ -199,22 +224,31 @@ static const GActionEntry dialog_actions[] =
   /* "Editor" tab */
   { MOUSEPAD_SETTING_INSERT_SPACES, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_AUTO_INDENT, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_INDENT_ON_TAB, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_SMART_BACKSPACE, mousepad_application_toggle_activate, NULL, "false", NULL },
 
   /* "Window" tab */
-  { MOUSEPAD_SETTING_STATUSBAR_VISIBLE, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_CLIENT_SIDE_DECORATIONS, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_PATH_IN_TITLE, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_REMEMBER_SIZE, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_REMEMBER_POSITION, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_REMEMBER_STATE, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_TOOLBAR_VISIBLE, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_ALWAYS_SHOW_TABS, mousepad_application_toggle_activate, NULL, "false", NULL },
-  { MOUSEPAD_SETTING_CYCLE_TABS, mousepad_application_toggle_activate, NULL, "false", NULL }
+  { MOUSEPAD_SETTING_EXPAND_TABS, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_CYCLE_TABS, mousepad_application_toggle_activate, NULL, "false", NULL },
+
+  /* "File" tab */
+  { MOUSEPAD_SETTING_ADD_LAST_EOL, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_MAKE_BACKUP, mousepad_application_toggle_activate, NULL, "false", NULL },
+  { MOUSEPAD_SETTING_MONITOR_CHANGES, mousepad_application_toggle_activate, NULL, "false", NULL }
 };
 #define N_DIALOG G_N_ELEMENTS (dialog_actions)
 
 /* settings only accessible from the menubar */
 static const GActionEntry menubar_actions[] =
 {
+  { MOUSEPAD_SETTING_SEARCH_INCREMENTAL, mousepad_application_toggle_activate, NULL, "false", NULL },
   { MOUSEPAD_SETTING_SEARCH_HIGHLIGHT_ALL, mousepad_application_toggle_activate, NULL, "false", NULL }
 };
 #define N_MENUBAR G_N_ELEMENTS (menubar_actions)
@@ -268,6 +302,10 @@ mousepad_application_class_init (MousepadApplicationClass *klass)
   application_class->open = mousepad_application_open;
   application_class->shutdown = mousepad_application_shutdown;
 
+  g_object_class_install_property (gobject_class, PROP_DEFAULT_FONT,
+    g_param_spec_string ("default-font", "DefaultFont", "The default font to use in text views",
+                         DEFAULT_FONT, G_PARAM_READWRITE));
+
   g_object_class_install_property (gobject_class, PROP_SPACE_LOCATION,
     g_param_spec_flags ("space-location", "SpaceLocation", "The space location setting",
                         GTK_SOURCE_TYPE_SPACE_LOCATION_FLAGS, GTK_SOURCE_SPACE_LOCATION_ALL,
@@ -286,6 +324,10 @@ mousepad_application_set_property (GObject      *object,
 
   switch (prop_id)
     {
+    case PROP_DEFAULT_FONT:
+      g_free (application->default_font);
+      application->default_font = g_value_dup_string (value);
+      break;
     case PROP_SPACE_LOCATION:
       application->space_location_flags = g_value_get_flags (value);
       break;
@@ -307,6 +349,9 @@ mousepad_application_get_property (GObject    *object,
 
   switch (prop_id)
     {
+    case PROP_DEFAULT_FONT:
+      g_value_set_string (value, application->default_font);
+      break;
     case PROP_SPACE_LOCATION:
       g_value_set_flags (value, application->space_location_flags);
       break;
@@ -328,11 +373,14 @@ mousepad_application_init (MousepadApplication *application)
 
   /* initialize application attributes */
   application->prefs_dialog = NULL;
+  application->prefs_dialog_standalone = FALSE;
+  application->default_font = g_strdup (DEFAULT_FONT);
   application->space_location_flags = GTK_SOURCE_SPACE_LOCATION_ALL;
   application->opening_mode = TAB;
   application->encoding = mousepad_encoding_get_default ();
   application->line = 0;
   application->column = 0;
+  application->providers = NULL;
 
   /* default application name */
   g_set_application_name (_("Mousepad"));
@@ -363,6 +411,7 @@ mousepad_application_parse_encoding (const gchar  *option_name,
                                      GError      **error)
 {
   MousepadApplication *application;
+  gboolean             user_set_encoding = TRUE;
 
   application = MOUSEPAD_APPLICATION (g_application_get_default ());
 
@@ -379,9 +428,13 @@ mousepad_application_parse_encoding (const gchar  *option_name,
         {
           g_printerr ("Invalid encoding '%s': ignored\n", value);
           application->encoding = mousepad_encoding_get_default ();
+          user_set_encoding = FALSE;
         }
     }
 
+  /* to be attached to the files to open later */
+  mousepad_object_set_data (application, "user-set-encoding", GINT_TO_POINTER (user_set_encoding));
+
   return TRUE;
 }
 
@@ -447,6 +500,8 @@ mousepad_application_handle_local_options (GApplication *gapplication,
 
   /* transfer encoding from remote to primary instance via the options dictionary */
   g_variant_dict_insert (options, "encoding", "u", application->encoding);
+  g_variant_dict_insert (options, "user-set-encoding", "b",
+    GPOINTER_TO_INT (mousepad_object_get_data (application, "user-set-encoding")));
 
   /* chain up to startup (primary instance) or command_line (remote instance) */
   return -1;
@@ -544,9 +599,12 @@ static void
 mousepad_application_set_accels (MousepadApplication *application)
 {
   GdkModifierType   accel_mods;
+  GTypeModule      *provider;
+  GList            *item;
   guint             n, accel_key;
   gchar            *accel_path;
   gchar           **action_names;
+  const gchar      *accel;
   const gchar      *accels[] = { NULL, NULL };
   const gchar      *accel_maps[][2] =
   {
@@ -568,8 +626,8 @@ mousepad_application_set_accels (MousepadApplication *application)
     { "win.edit.delete-line", "<Control><Shift>Delete" }, { "win.edit.select-all", "<Control>A" },
     { "win.edit.convert.to-opposite-case", "<Alt><Control>U" },
     { "win.edit.convert.transpose", "<Control>T" },
-    { "win.edit.move-selection.line-up", "<Alt>Page_Up" },
-    { "win.edit.move-selection.line-down", "<Alt>Page_Down" },
+    { "win.edit.move.line-up", "<Alt>Up" }, { "win.edit.move.line-down", "<Alt>Down" },
+    { "win.edit.move.word-left", "<Alt>Left" }, { "win.edit.move.word-right", "<Alt>Right" },
     { "win.edit.increase-indent", "<Control>I" }, { "win.edit.decrease-indent", "<Control>U" },
 
     /* "Search" menu */
@@ -609,6 +667,25 @@ mousepad_application_set_accels (MousepadApplication *application)
                                              accel_maps[n][0], accels);
     }
 
+  /* plugin enabling actions */
+  for (item = application->providers; item != NULL; item = item->next)
+    {
+      if ((accel = mousepad_plugin_provider_get_accel (item->data)) == NULL)
+        continue;
+
+      /* add accel map entry to fill the accels file at shutdown */
+      provider = item->data;
+      accel_path = g_strconcat ("<Actions>/app.", provider->name, NULL);
+      gtk_accelerator_parse (accel, &accel_key, &accel_mods);
+      gtk_accel_map_add_entry (accel_path, accel_key, accel_mods);
+
+      /* add accel to the application */
+      accels[0] = accel;
+      gtk_application_set_accels_for_action (GTK_APPLICATION (application),
+                                             accel_path + 10, accels);
+      g_free (accel_path);
+    }
+
   /* actions that do not have a default accel */
   action_names = g_action_group_list_actions (G_ACTION_GROUP (application));
   n = 0;
@@ -632,10 +709,137 @@ mousepad_application_set_accels (MousepadApplication *application)
 
 
 
+static gint
+mousepad_application_sort_plugins (gconstpointer a,
+                                   gconstpointer b)
+{
+  MousepadPluginProvider *p = MOUSEPAD_PLUGIN_PROVIDER (a),
+                         *q = MOUSEPAD_PLUGIN_PROVIDER (b);
+  gint comp_cat;
+
+  comp_cat = g_utf8_collate (mousepad_plugin_provider_get_category (p),
+                             mousepad_plugin_provider_get_category (q));
+
+  /* alphabetical sorting of categories, and within categories of labels */
+  if (comp_cat == 0)
+    return g_utf8_collate (mousepad_plugin_provider_get_label (p),
+                           mousepad_plugin_provider_get_label (q));
+  else
+    return comp_cat;
+}
+
+
+
+static void
+mousepad_application_load_plugins (MousepadApplication *application)
+{
+  MousepadPluginProvider  *provider;
+  GSimpleAction           *action;
+  GError                  *error = NULL;
+  GDir                    *dir;
+  const gchar             *basename;
+  gchar                   *provider_name, *schema_id;
+  gchar                  **strs;
+  gsize                    n_strs;
+
+  if (! g_module_supported ())
+    {
+      g_warning ("Dynamic type loading is not supported on this system");
+      return;
+    }
+  else if ((dir = g_dir_open (MOUSEPAD_PLUGIN_DIRECTORY, 0, &error)) == NULL)
+    {
+      /* the plugin directory may not exist (compilation without plugin) */
+      if (g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+        g_message ("Plugin directory '%s' not found", MOUSEPAD_PLUGIN_DIRECTORY);
+      else
+        g_warning ("Failed to open plugin directory '%s': %s",
+                   MOUSEPAD_PLUGIN_DIRECTORY, error->message);
+
+      return;
+    }
+
+  /* scan plugin directory */
+  for (basename = g_dir_read_name (dir); basename != NULL; basename = g_dir_read_name (dir))
+    {
+      if (! g_str_has_prefix (basename, "lib")
+          || ! g_str_has_suffix (basename, "." G_MODULE_SUFFIX))
+        continue;
+
+      /* remove prefix */
+      provider_name = (gchar *) (basename + 3);
+
+      /* remove suffix */
+      strs = g_strsplit (provider_name, ".", -1);
+      n_strs = g_strv_length (strs);
+      g_free (strs[n_strs - 1]);
+      strs[n_strs - 1] = NULL;
+      provider_name = g_strjoinv (".", strs);
+      g_strfreev (strs);
+
+      /* get the list of enabled plugins */
+      strs = MOUSEPAD_SETTING_GET_STRV (ENABLED_PLUGINS);
+
+      /* try to load this module */
+      provider = mousepad_plugin_provider_new (provider_name);
+      if (g_type_module_use (G_TYPE_MODULE (provider)))
+        {
+          /* add this provider to the list */
+          application->providers = g_list_prepend (application->providers, provider);
+
+          /* create its action and add it to the application */
+          action = g_simple_action_new_stateful (provider_name, NULL,
+                                                 g_variant_new_boolean (FALSE));
+          g_signal_connect (action, "activate",
+                            G_CALLBACK (mousepad_application_plugin_activate), application);
+          MOUSEPAD_SETTING_CONNECT_OBJECT (ENABLED_PLUGINS, mousepad_application_plugin_update,
+                                           application, G_CONNECT_SWAPPED);
+          g_action_map_add_action (G_ACTION_MAP (application), G_ACTION (action));
+
+          /* add its settings to the setting store */
+          if (g_str_has_prefix (provider_name, "mousepad-plugin-"))
+            schema_id = provider_name + 16;
+          else
+            schema_id = provider_name;
+
+          schema_id = g_strconcat (MOUSEPAD_ID, ".plugins.", schema_id, NULL);
+          mousepad_settings_add_root (schema_id);
+          g_free (schema_id);
+
+          /* instantiate this provider types and initialize its action state */
+          if (g_strv_contains ((const gchar *const *) strs, provider_name))
+            {
+              mousepad_plugin_provider_new_plugin (provider);
+              g_simple_action_set_state (action, g_variant_new_boolean (TRUE));
+            }
+          else
+            g_type_module_unuse (G_TYPE_MODULE (provider));
+        }
+      else
+        g_object_unref (provider);
+
+      /* cleanup */
+      g_strfreev (strs);
+      g_free (provider_name);
+    }
+
+  /* ends scan */
+  g_dir_close (dir);
+
+  /* sort the list */
+  application->providers = g_list_sort (application->providers,
+                                        mousepad_application_sort_plugins);
+}
+
+
+
 static void
 mousepad_application_startup (GApplication *gapplication)
 {
+  static GSettings *settings;
+
   MousepadApplication *application = MOUSEPAD_APPLICATION (gapplication);
+  GSettingsSchema     *schema;
   GVariant            *state;
   GAction             *action;
   GMenu               *menu;
@@ -644,6 +848,24 @@ mousepad_application_startup (GApplication *gapplication)
   /* chain up to parent */
   G_APPLICATION_CLASS (mousepad_application_parent_class)->startup (gapplication);
 
+  /* load plugins */
+  mousepad_application_load_plugins (application);
+
+  /* bind the default font to GNOME settings if possible */
+  schema = g_settings_schema_source_lookup (g_settings_schema_source_get_default (),
+                                            "org.gnome.desktop.interface", TRUE);
+  if (schema != NULL)
+    {
+      if (g_settings_schema_has_key (schema, "monospace-font-name"))
+        {
+          settings = g_settings_new ("org.gnome.desktop.interface");
+          g_settings_bind (settings, "monospace-font-name",
+                           application, "default-font", G_SETTINGS_BIND_GET);
+        }
+
+      g_settings_schema_unref (schema);
+    }
+
   /* add application actions */
   g_action_map_add_action_entries (G_ACTION_MAP (application), stateless_actions,
                                    N_STATELESS, application);
@@ -699,6 +921,9 @@ mousepad_application_startup (GApplication *gapplication)
   /* do some actions when the active window changes */
   g_signal_connect (application, "notify::active-window",
                     G_CALLBACK (mousepad_application_active_window_changed), NULL);
+
+  /* initialize history management */
+  mousepad_history_init ();
 }
 
 
@@ -709,26 +934,45 @@ mousepad_application_command_line (GApplication            *gapplication,
 {
   MousepadApplication  *application = MOUSEPAD_APPLICATION (gapplication);
   GVariantDict         *options;
-  GError               *error = NULL;
-  GPtrArray            *files;
+  GFile               **files;
   GFile                *file;
-  gpointer             *data;
-  const gchar          *opening_mode, *working_directory;
-  gchar               **filenames = NULL;
+  const gchar          *opening_mode;
+  const gchar         **filenames = NULL;
   gint                  n, n_files;
-
-  /* initialize xfconf */
-  if (G_UNLIKELY (xfconf_init (&error) == FALSE))
-    {
-      g_critical ("Failed to initialize xfconf: %s", error->message);
-      g_error_free (error);
-
-      return EXIT_FAILURE;
-    }
+  gboolean              user_set_encoding, user_set_cursor = FALSE, restored = FALSE;
 
   /* get the option dictionary */
   options = g_application_command_line_get_options_dict (command_line);
 
+  /* see if the prefs dialog is to be opened */
+  if (g_variant_dict_contains (options, "preferences"))
+    {
+      /* create and show the prefs dialog */
+      g_action_group_activate_action (G_ACTION_GROUP (gapplication), "preferences", NULL);
+
+      /* increase application use count to keep the prefs dialog alive */
+      if (! application->prefs_dialog_standalone)
+        {
+          g_application_hold (gapplication);
+          application->prefs_dialog_standalone = TRUE;
+        }
+
+      return EXIT_SUCCESS;
+    }
+
+  /* restore previous session */
+  if (MOUSEPAD_SETTING_GET_ENUM (SESSION_RESTORE) != MOUSEPAD_SESSION_RESTORE_NEVER
+      && ! g_application_command_line_get_is_remote (command_line))
+    {
+      application->opening_mode = MIXED;
+      application->encoding = mousepad_encoding_get_default ();
+      restored = mousepad_history_session_restore (application);
+    }
+
+  /* retrieve encoding from the remote instance */
+  g_variant_dict_lookup (options, "encoding", "u", &(application->encoding));
+  g_variant_dict_lookup (options, "user-set-encoding", "b", &user_set_encoding);
+
   /* see if an opening mode was provided on the command line */
   if (g_variant_dict_lookup (options, "opening-mode", "&s", &opening_mode))
     {
@@ -750,45 +994,54 @@ mousepad_application_command_line (GApplication            *gapplication,
   else
     application->opening_mode = MOUSEPAD_SETTING_GET_ENUM (OPENING_MODE);
 
-  /* retrieve encoding from the remote instance */
-  g_variant_dict_lookup (options, "encoding", "u", &(application->encoding));
+  /* see if line number was not provided on the command line */
+  if (! g_variant_dict_lookup (options, "line", "i", &(application->line)))
+    application->line = 0;
+  else
+    {
+      user_set_cursor = TRUE;
 
-  /* get line and column number from command line */
-  g_variant_dict_lookup (options, "line", "i", &(application->line));
-  g_variant_dict_lookup (options, "column", "i", &(application->column));
+      /* for user line starts from 1 but for gtk line starts from 0 */
+      if (application->line > 0)
+        --application->line;
+    }
 
-  /* for user line starts from 1 but for gtk line starts from 0 */
-  if (application->line > 0)
-    --application->line;
+  /* see if column number was not provided on the command line */
+  if (! g_variant_dict_lookup (options, "column", "i", &(application->column)))
+    application->column = 0;
+  else
+    user_set_cursor = TRUE;
 
   /* extract filenames */
   g_variant_dict_lookup (options, G_OPTION_REMAINING, "^a&ay", &filenames);
 
-  /* get the current working directory */
-  working_directory = g_application_command_line_get_cwd (command_line);
-
-  /* open files provided on the command line or an empty document */
-  if (working_directory && filenames && (n_files = g_strv_length (filenames)))
+  /* open files provided on the command line */
+  if (filenames != NULL && (n_files = g_strv_length ((gchar **) filenames)) > 0)
     {
-      /* prepare the GFiles array */
-      files = g_ptr_array_new ();
+      /* prepare the GFile array */
+      files = g_malloc (n_files * sizeof (GFile *));
       for (n = 0; n < n_files; n++)
         {
           file = g_application_command_line_create_file_for_arg (command_line, filenames[n]);
-          g_ptr_array_add (files, file);
+          mousepad_object_set_data (file, "user-set-encoding",
+                                    GINT_TO_POINTER (user_set_encoding));
+          mousepad_object_set_data (file, "user-set-cursor",
+                                    GINT_TO_POINTER (user_set_cursor));
+          files[n] = file;
         }
-      data = g_ptr_array_free (files, FALSE);
 
       /* open the files */
-      mousepad_application_open (gapplication, (GFile **) data, n_files, NULL);
+      g_application_open (gapplication, files, n_files, NULL);
 
       /* cleanup */
       for (n = 0; n < n_files; n++)
-        g_object_unref (data[n]);
-      g_free (data);
+        g_object_unref (files[n]);
+
+      g_free (files);
     }
-  else
-    mousepad_application_activate (gapplication);
+  /* open an empty document if previous session wasn't restored */
+  else if (! restored)
+    g_application_activate (gapplication);
 
   /* cleanup */
   g_free (filenames);
@@ -827,8 +1080,8 @@ mousepad_application_activate (GApplication *gapplication)
   /* create a new document and add it to the window */
   g_action_group_activate_action (G_ACTION_GROUP (window), "file.new", NULL);
 
-  /* show the window */
-  gtk_widget_show (window);
+  /* present the window, so that it requires attention if already open */
+  gtk_window_present (GTK_WINDOW (window));
 }
 
 
@@ -847,7 +1100,7 @@ mousepad_application_open (GApplication  *gapplication,
 
   /* URIs may be invalid when entered by the user */
   for (n = 0; n < n_files; n++)
-    if (mousepad_util_get_path (files[n]) == NULL)
+    if (! mousepad_util_validate_file (files[n]))
       {
         /* inform the user */
         uri = g_file_get_uri (files[n]);
@@ -874,7 +1127,8 @@ mousepad_application_open (GApplication  *gapplication,
            * attention if already open */
           if (opened > 0)
             gtk_window_present (GTK_WINDOW (window));
-          else if (GTK_IS_WIDGET (window))
+          /* destroy the window if it was not already destroyed, e.g. by "app.quit" */
+          else if (G_LIKELY (mousepad_is_application_window (window)) && opened < 0)
             gtk_widget_destroy (window);
         }
       /* open the files in windows */
@@ -893,7 +1147,8 @@ mousepad_application_open (GApplication  *gapplication,
               /* if the file was finally opened, show the window */
               if (opened > 0)
                 gtk_widget_show (window);
-              else if (GTK_IS_WIDGET (window))
+              /* destroy the window if it was not already destroyed, e.g. by "app.quit" */
+              else if (G_LIKELY (mousepad_is_application_window (window)))
                 gtk_widget_destroy (window);
             }
         }
@@ -909,21 +1164,13 @@ mousepad_application_shutdown (GApplication *gapplication)
   GList               *windows, *window;
   gchar               *filename;
 
-  /* save the current accel map */
-  filename = mousepad_util_get_save_location (MOUSEPAD_ACCELS_RELPATH, TRUE);
-  if (G_LIKELY (filename != NULL))
-    {
-      gtk_accel_map_save (filename);
-      g_free (filename);
-    }
-
   /* flush the history items of the replace dialog
    * this is a bit of an ugly place, but cleaning on a window close
    * isn't a good option eighter */
   mousepad_replace_dialog_history_clean ();
 
   /* destroy the preferences dialog */
-  if (GTK_IS_WIDGET (application->prefs_dialog))
+  if (application->prefs_dialog != NULL)
     gtk_widget_destroy (application->prefs_dialog);
 
   /* destroy the windows if they are still opened */
@@ -933,12 +1180,28 @@ mousepad_application_shutdown (GApplication *gapplication)
 
   g_list_free (windows);
 
+  /* unload plugins */
+  g_list_free_full (application->providers, mousepad_plugin_provider_unuse);
+
+  /* release property related variables */
+  g_free (application->default_font);
+
+  /* save the current accel map */
+  filename = mousepad_util_get_save_location (MOUSEPAD_ACCELS_RELPATH, TRUE);
+  if (G_LIKELY (filename != NULL))
+    {
+      gtk_accel_map_save (filename);
+      g_free (filename);
+    }
+
+  /* we're about to quit, the user uses session restore only on crash and everything
+   * went well: clear session array */
+  if (MOUSEPAD_SETTING_GET_ENUM (SESSION_RESTORE) == MOUSEPAD_SESSION_RESTORE_CRASH)
+    MOUSEPAD_SETTING_RESET (SESSION);
+
   /* finalize mousepad settings */
   mousepad_settings_finalize ();
 
-  /* shutdown xfconf */
-  xfconf_shutdown ();
-
   /* chain up to parent */
   G_APPLICATION_CLASS (mousepad_application_parent_class)->shutdown (gapplication);
 }
@@ -949,7 +1212,7 @@ static GtkWidget *
 mousepad_application_create_window (MousepadApplication *application)
 {
   GtkWindowGroup *window_group;
-  GtkWidget      *window;
+  GtkWidget      *window, *notebook;
 
   /* create a new window */
   window = mousepad_window_new (application);
@@ -967,6 +1230,11 @@ mousepad_application_create_window (MousepadApplication *application)
                     G_CALLBACK (mousepad_application_new_window_with_document), application);
   g_signal_connect (window, "new-window",
                     G_CALLBACK (mousepad_application_new_window), application);
+  notebook = mousepad_window_get_notebook (MOUSEPAD_WINDOW (window));
+  g_signal_connect_after (notebook, "switch-page",
+                          G_CALLBACK (mousepad_history_session_save), NULL);
+  g_signal_connect_after (notebook, "page-reordered",
+                          G_CALLBACK (mousepad_history_session_save), NULL);
 
   return window;
 }
@@ -1036,10 +1304,13 @@ mousepad_application_active_window_changed (MousepadApplication *application)
       && g_list_find (windows, app_windows->data))
     {
       /* update document dependent menu items */
-      mousepad_window_update_document_menu_items (MOUSEPAD_WINDOW (app_windows->data));
+      mousepad_window_update_document_menu_items (app_windows->data);
 
       /* update window dependent menu items */
-      mousepad_window_update_window_menu_items (MOUSEPAD_WINDOW (app_windows->data));
+      mousepad_window_update_window_menu_items (app_windows->data);
+
+      /* save new session state */
+      mousepad_history_session_save ();
     }
 
   /* store a copy of the application windows list to compare at next call */
@@ -1049,54 +1320,6 @@ mousepad_application_active_window_changed (MousepadApplication *application)
 
 
 
-static void
-mousepad_application_prefs_dialog_response (MousepadApplication *application,
-                                            gint                 response_id,
-                                            MousepadPrefsDialog *dialog)
-{
-  gtk_widget_destroy (application->prefs_dialog);
-  application->prefs_dialog = NULL;
-}
-
-
-
-void
-mousepad_application_show_preferences (MousepadApplication *application,
-                                       GtkWindow           *transient_for)
-{
-  GList *windows;
-
-  /* if the dialog isn't already shown, create one */
-  if (! GTK_IS_WIDGET (application->prefs_dialog))
-    {
-      application->prefs_dialog = mousepad_prefs_dialog_new ();
-
-      /* destroy the dialog when it's close button is pressed */
-      g_signal_connect_swapped (application->prefs_dialog, "response",
-                                G_CALLBACK (mousepad_application_prefs_dialog_response),
-                                application);
-    }
-
-  /* if no transient window was specified, used the first application window
-   * or NULL if no windows exists (shouldn't happen) */
-  if (! GTK_IS_WINDOW (transient_for))
-    {
-      windows = gtk_application_get_windows (GTK_APPLICATION (application));
-      if (windows && GTK_IS_WINDOW (windows->data))
-        transient_for = GTK_WINDOW (windows->data);
-      else
-        transient_for = NULL;
-    }
-
-  /* associate it with one of the windows */
-  gtk_window_set_transient_for (GTK_WINDOW (application->prefs_dialog), transient_for);
-
-  /* show it to the user */
-  gtk_window_present (GTK_WINDOW (application->prefs_dialog));
-}
-
-
-
 static void
 mousepad_application_update_menu (GMenuModel *shared_menu,
                                   gint        position,
@@ -1336,30 +1559,91 @@ mousepad_application_create_style_schemes_menu (MousepadApplication *application
 
 
 
+static void
+mousepad_application_prefs_dialog_standalone (MousepadApplication *application)
+{
+  if (application->prefs_dialog_standalone)
+    {
+      g_application_release (G_APPLICATION (application));
+      application->prefs_dialog_standalone = FALSE;
+    }
+}
+
+
+
+static void
+mousepad_application_prefs_dialog_response (MousepadApplication *application,
+                                            gint                 response_id,
+                                            MousepadPrefsDialog *dialog)
+{
+  gtk_widget_destroy (application->prefs_dialog);
+  application->prefs_dialog = NULL;
+
+  /* decrease application use count, if needed */
+  mousepad_application_prefs_dialog_standalone (application);
+}
+
+
+
+static void
+mousepad_application_action_preferences (GSimpleAction *action,
+                                         GVariant      *value,
+                                         gpointer       data)
+{
+  MousepadApplication *application = data;
+
+  /* if the dialog isn't already shown, create one */
+  if (application->prefs_dialog == NULL)
+    {
+      application->prefs_dialog = mousepad_prefs_dialog_new ();
+
+      /* destroy the dialog when it's close button is pressed */
+      g_signal_connect_swapped (application->prefs_dialog, "response",
+                                G_CALLBACK (mousepad_application_prefs_dialog_response),
+                                application);
+    }
+
+  /* associate it with the active window, if any */
+  gtk_window_set_transient_for (GTK_WINDOW (application->prefs_dialog),
+                                gtk_application_get_active_window (data));
+
+  /* show it to the user */
+  gtk_window_present (GTK_WINDOW (application->prefs_dialog));
+}
+
+
+
 static void
 mousepad_application_action_quit (GSimpleAction *action,
                                   GVariant      *value,
                                   gpointer       data)
 {
-  GList    *windows, *window;
-  GAction  *close_action;
-  GVariant *v_succeed;
-  gboolean  succeed;
+  GList   *windows, *window;
+  GAction *close;
+
+  /* block session handler */
+  mousepad_history_session_set_quitting (TRUE);
 
   /* try to close all windows, abort at the first failure */
-  windows = g_list_copy (gtk_application_get_windows (GTK_APPLICATION (data)));
+  windows = g_list_copy (gtk_application_get_windows (data));
   for (window = windows; window != NULL; window = window->next)
     {
-      close_action = g_action_map_lookup_action (G_ACTION_MAP (window->data), "file.close-window");
-      g_action_activate (close_action, NULL);
-      v_succeed = g_action_get_state (close_action);
-      succeed = g_variant_get_int32 (v_succeed);
-      g_variant_unref (v_succeed);
-      if (! succeed)
-        break;
+      close = g_action_map_lookup_action (G_ACTION_MAP (window->data), "file.close-window");
+      g_action_activate (close, NULL);
+      if (! mousepad_action_get_state_int32_boolean (close))
+        {
+          /* unblock session handler and save session */
+          mousepad_history_session_set_quitting (FALSE);
+          mousepad_history_session_save ();
+
+          break;
+        }
     }
 
   g_list_free (windows);
+
+  /* decrease application use count, if needed */
+  mousepad_application_prefs_dialog_standalone (data);
 }
 
 
@@ -1372,7 +1656,7 @@ mousepad_application_toggle_activate (GSimpleAction *action,
   gboolean state;
 
   /* save the setting */
-  state = ! g_variant_get_boolean (g_action_get_state (G_ACTION (action)));
+  state = ! mousepad_action_get_state_boolean (G_ACTION (action));
   mousepad_setting_set_boolean (g_action_get_name (G_ACTION (action)), state);
 }
 
@@ -1389,6 +1673,56 @@ mousepad_application_radio_activate (GSimpleAction *action,
 
 
 
+static void
+mousepad_application_plugin_activate (GSimpleAction *action,
+                                      GVariant      *parameter,
+                                      gpointer       data)
+{
+  gchar       **plugins;
+  const gchar  *action_name;
+  gboolean      enabled, contained, need_update = FALSE;
+  guint         length = 0;
+
+  /* get the new action state */
+  enabled = ! mousepad_action_get_state_boolean (G_ACTION (action));
+
+  /* get the list of enabled plugins */
+  plugins = MOUSEPAD_SETTING_GET_STRV (ENABLED_PLUGINS);
+
+  /* check if this plugin is in the list */
+  action_name = g_action_get_name (G_ACTION (action));
+  contained = g_strv_contains ((const gchar *const *) plugins, action_name);
+
+  /* update the list if needed */
+  if (enabled && ! contained)
+    {
+      need_update = TRUE;
+      length = g_strv_length (plugins);
+      plugins = g_realloc (plugins, (length + 2) * sizeof (gchar *));
+      plugins[length] = g_strdup (action_name);
+      plugins[length + 1] = NULL;
+    }
+  else if (! enabled && contained)
+    {
+      need_update = TRUE;
+      while (g_strcmp0 (plugins[length++], action_name) != 0);
+      g_free (plugins[--length]);
+      while (plugins[++length] != NULL)
+        plugins[length - 1] = plugins[length];
+
+      plugins[length - 1] = NULL;
+    }
+
+  /* save the setting */
+  if (need_update)
+    MOUSEPAD_SETTING_SET_STRV (ENABLED_PLUGINS, (const gchar *const *) plugins);
+
+  /* cleanup */
+  g_strfreev (plugins);
+}
+
+
+
 static void
 mousepad_application_action_update (MousepadApplication *application,
                                     gchar               *key,
@@ -1413,6 +1747,49 @@ mousepad_application_action_update (MousepadApplication *application,
 
 
 
+static void
+mousepad_application_plugin_update (MousepadApplication *application)
+{
+  MousepadPluginProvider  *mprovider;
+  GTypeModule             *provider;
+  GAction                 *action;
+  GList                   *item;
+  gchar                  **plugins;
+  gboolean                 enabled, contained, destroyable;
+
+  /* get the list of enabled plugins */
+  plugins = MOUSEPAD_SETTING_GET_STRV (ENABLED_PLUGINS);
+
+  /* check for each plugin if its action state should be changed, and if so,
+   * if we should do something on the plugin itself from here */
+  for (item = application->providers; item != NULL; item = item->next)
+    {
+      provider = item->data;
+      contained = g_strv_contains ((const gchar *const *) plugins, provider->name);
+      action = g_action_map_lookup_action (G_ACTION_MAP (application), provider->name);
+      enabled = mousepad_action_get_state_boolean (action);
+      if ((enabled && ! contained) || (! enabled && contained))
+        {
+          g_action_change_state (action, g_variant_new_boolean (! enabled));
+          mprovider = item->data;
+          destroyable = mousepad_plugin_provider_is_destroyable (mprovider);
+
+          /* plugin should be loaded from here */
+          if (! enabled && (! mousepad_plugin_provider_is_instantiated (mprovider) || destroyable)
+              && g_type_module_use (provider))
+            mousepad_plugin_provider_new_plugin (mprovider);
+          /* plugin should be unloaded from here */
+          else if (enabled && destroyable)
+            g_type_module_unuse (provider);
+        }
+    }
+
+  /* cleanup */
+  g_strfreev (plugins);
+}
+
+
+
 /*
  * This job is done only once here, when whitespace location settings are updated,
  * then all the views update their own location flags from those of the application.
@@ -1435,3 +1812,11 @@ mousepad_application_action_whitespace (GSimpleAction *action,
   /* set the application property */
   g_object_set (data, "space-location", flags, NULL);
 }
+
+
+
+GList *
+mousepad_application_get_providers (MousepadApplication *application)
+{
+  return application->providers;
+}
