@@ -20,6 +20,7 @@
 #include <mousepad/mousepad-encoding-dialog.h>
 #include <mousepad/mousepad-settings.h>
 #include <mousepad/mousepad-window.h>
+#include <mousepad/mousepad-history.h>
 
 
 
@@ -89,9 +90,12 @@ mousepad_dialogs_show_error (GtkWindow    *parent,
 
   /* create the warning dialog */
   dialog = gtk_message_dialog_new (parent, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR,
-                                   GTK_BUTTONS_CLOSE, "%s.", message);
+                                   GTK_BUTTONS_NONE, "%s.", message);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* set secondary text if an error is provided */
   if (G_LIKELY (error != NULL))
     gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s.", error->message);
@@ -134,10 +138,13 @@ mousepad_dialogs_other_tab_size (GtkWindow *parent,
 
   /* build dialog */
   dialog = gtk_dialog_new_with_buttons (_("Select Tab Size"), parent, GTK_DIALOG_MODAL,
-                                        _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL,
-                                        _("_OK"), MOUSEPAD_RESPONSE_OK, NULL);
+                                        MOUSEPAD_LABEL_CANCEL, MOUSEPAD_RESPONSE_CANCEL,
+                                        MOUSEPAD_LABEL_OK, MOUSEPAD_RESPONSE_OK, NULL);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* set properties */
   gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_OK);
 
@@ -169,6 +176,7 @@ mousepad_dialogs_go_to_line_changed (GtkSpinButton *line_spin,
 {
   GtkTextBuffer *buffer;
   GtkTextIter    iter;
+  gint           line, total_columns;
 
   g_return_if_fail (GTK_IS_SPIN_BUTTON (line_spin));
   g_return_if_fail (GTK_IS_SPIN_BUTTON (col_spin));
@@ -176,15 +184,24 @@ mousepad_dialogs_go_to_line_changed (GtkSpinButton *line_spin,
   /* get the text buffer */
   buffer = mousepad_object_get_data (col_spin, "buffer");
 
+  line = gtk_spin_button_get_value_as_int (line_spin);
+
+  if (line > 0)
+    line--;
+  else if (line < 0)
+    line += gtk_text_buffer_get_line_count (buffer);
+
   /* get iter at line */
-  gtk_text_buffer_get_iter_at_line (buffer, &iter, gtk_spin_button_get_value_as_int (line_spin) - 1);
+  gtk_text_buffer_get_iter_at_line (buffer, &iter, line);
 
   /* move the iter to the end of the line if needed */
   if (!gtk_text_iter_ends_line (&iter))
     gtk_text_iter_forward_to_line_end (&iter);
 
+  total_columns = mousepad_util_get_real_line_offset (&iter);
+
   /* update column spin button range */
-  gtk_spin_button_set_range (col_spin, 0, gtk_text_iter_get_line_offset (&iter));
+  gtk_spin_button_set_range (col_spin, -(total_columns + 1), total_columns);
 }
 
 
@@ -206,15 +223,19 @@ mousepad_dialogs_go_to (GtkWindow     *parent,
   /* get cursor iter */
   gtk_text_buffer_get_iter_at_mark (buffer, &iter, gtk_text_buffer_get_insert (buffer));
   line = gtk_text_iter_get_line (&iter) + 1;
+  column = mousepad_util_get_real_line_offset (&iter);
 
   /* get number of lines */
   lines = gtk_text_buffer_get_line_count (buffer);
 
   /* build the dialog */
   dialog = gtk_dialog_new_with_buttons (_("Go To"), parent, GTK_DIALOG_MODAL,
-                                        _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL, NULL);
+                                        MOUSEPAD_LABEL_CANCEL, MOUSEPAD_RESPONSE_CANCEL, NULL);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* add button */
   button = mousepad_util_image_button ("go-jump", _("_Jump to"));
   gtk_widget_set_can_default (button, TRUE);
@@ -243,7 +264,7 @@ mousepad_dialogs_go_to (GtkWindow     *parent,
   gtk_label_set_yalign (GTK_LABEL (label), 0.5);
   gtk_widget_show (label);
 
-  line_spin = gtk_spin_button_new_with_range (1, lines, 1);
+  line_spin = gtk_spin_button_new_with_range (-lines, lines, 1);
   gtk_entry_set_activates_default (GTK_ENTRY (line_spin), TRUE);
   gtk_box_pack_start (GTK_BOX (hbox), line_spin, FALSE, FALSE, 0);
   gtk_label_set_mnemonic_widget (GTK_LABEL (label), line_spin);
@@ -278,7 +299,9 @@ mousepad_dialogs_go_to (GtkWindow     *parent,
   /* signal to monitor column number */
   g_signal_connect (line_spin, "value-changed",
                     G_CALLBACK (mousepad_dialogs_go_to_line_changed), col_spin);
+
   gtk_spin_button_set_value (GTK_SPIN_BUTTON (line_spin), line);
+  gtk_spin_button_set_value (GTK_SPIN_BUTTON (col_spin), column);
 
   /* run the dialog */
   response = gtk_dialog_run (GTK_DIALOG (dialog));
@@ -288,14 +311,13 @@ mousepad_dialogs_go_to (GtkWindow     *parent,
       gtk_widget_hide (dialog);
 
       /* get new position */
-      line = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (line_spin)) - 1;
+      line = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (line_spin));
+      if (line > 0)
+        line--;
       column = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (col_spin));
 
-      /* get iter */
-      gtk_text_buffer_get_iter_at_line_offset (buffer, &iter, line, column);
-
-      /* get cursor position */
-      gtk_text_buffer_place_cursor (buffer, &iter);
+      /* place cursor at (line, column) */
+      mousepad_util_place_cursor (buffer, line, column);
     }
 
   /* destroy the dialog */
@@ -309,53 +331,30 @@ mousepad_dialogs_go_to (GtkWindow     *parent,
 gboolean
 mousepad_dialogs_clear_recent (GtkWindow *parent)
 {
-  GtkWidget *dialog;
-  GtkWidget *area, *hbox, *vbox;
-  GtkWidget *button;
-  GtkWidget *image;
-  GtkWidget *label;
+  GtkWidget *dialog, *button;
   gboolean   succeed = FALSE;
 
-  /* create the question dialog */
-  dialog = gtk_dialog_new_with_buttons (_("Clear Documents History"), parent, GTK_DIALOG_MODAL,
-                                        _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL, NULL);
+  /* setup the question dialog */
+  dialog = gtk_message_dialog_new (parent, GTK_DIALOG_MODAL,
+                                   GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
+                                   _("Remove all entries from the documents history?"));
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
-  /* set button */
+  /* setup CSD titlebar */
+  gtk_window_set_title (GTK_WINDOW (dialog), _("Clear Documents History"));
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
+  /* set secondary text */
+  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+    _("Clearing the documents history will permanently remove all currently listed entries."));
+
+  /* add buttons */
+  gtk_dialog_add_buttons (GTK_DIALOG (dialog), MOUSEPAD_LABEL_CANCEL,
+                          MOUSEPAD_RESPONSE_CANCEL, NULL);
+  gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_CANCEL);
+
   button = mousepad_util_image_button ("edit-clear", _("Clea_r"));
   gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_CLEAR);
-  gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_CANCEL);
-  gtk_window_set_default_size (GTK_WINDOW (dialog), 400, -1);
-
-  /* the content area */
-  area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
-  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 8);
-  gtk_box_pack_start (GTK_BOX (area), hbox, TRUE, TRUE, 6);
-  gtk_widget_show (hbox);
-
-  /* the dialog icon */
-  image = gtk_image_new_from_icon_name ("edit-clear", GTK_ICON_SIZE_DIALOG);
-  gtk_box_pack_start (GTK_BOX (hbox), image, TRUE, TRUE, 0);
-  gtk_widget_show (image);
-
-  /* the dialog message */
-  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 8);
-  gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 6);
-  gtk_widget_show (vbox);
-
-  /* primary text */
-  label = gtk_label_new (NULL);
-  gtk_label_set_markup (GTK_LABEL (label),
-                        _("<big><b>Remove all entries from the documents history?</b></big>"));
-  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, TRUE, 0);
-  gtk_widget_show (label);
-
-  /* secondary text */
-  label = gtk_label_new (_("Clearing the documents history will permanently "
-                           "remove all currently listed entries."));
-  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
-  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, TRUE, 0);
-  gtk_widget_show (label);
 
   /* popup the dialog */
   if (gtk_dialog_run (GTK_DIALOG (dialog)) == MOUSEPAD_RESPONSE_CLEAR)
@@ -371,70 +370,65 @@ mousepad_dialogs_clear_recent (GtkWindow *parent)
 
 gint
 mousepad_dialogs_save_changes (GtkWindow *parent,
+                               gboolean   closing,
                                gboolean   readonly)
 {
-  GtkWidget *dialog;
-  GtkWidget *area, *hbox, *vbox;
-  GtkWidget *button;
-  GtkWidget *image;
-  GtkWidget *label;
-  gint       response;
+  GtkWidget   *dialog, *button;
+  const gchar *text;
+  gint         response, button_response;
+
+  /* primary text */
+  if (G_LIKELY (closing))
+    text = _("Do you want to save the changes before closing?");
+  else
+    text = _("The document is read-only, do you want to save it as another file?");
 
   /* create the question dialog */
-  dialog = gtk_dialog_new_with_buttons (_("Save Changes"), parent, GTK_DIALOG_MODAL,
-                                        _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL, NULL);
+  dialog = gtk_message_dialog_new_with_markup (parent, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION,
+                                               GTK_BUTTONS_NONE, "<b><big>%s</big></b>", text);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
-  /* set properties */
-  gtk_window_set_default_size (GTK_WINDOW (dialog), 400, -1);
-  gtk_dialog_add_action_widget (GTK_DIALOG (dialog),
-                                mousepad_util_image_button ("edit-delete", _("_Don't Save")),
-                                MOUSEPAD_RESPONSE_DONT_SAVE);
+  /* setup CSD titlebar */
+  gtk_window_set_title (GTK_WINDOW (dialog), _("Save Changes"));
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
 
-  /* we show the save as button instead of save for readonly document */
-  if (G_UNLIKELY (readonly))
+  /* add buttons */
+  gtk_dialog_add_buttons (GTK_DIALOG (dialog), MOUSEPAD_LABEL_CANCEL,
+                          MOUSEPAD_RESPONSE_CANCEL, NULL);
+
+  if (G_LIKELY (closing))
     {
-      image = gtk_image_new_from_icon_name ("document-save-as", GTK_ICON_SIZE_DIALOG);
-      button = mousepad_util_image_button ("document-save-as", _("_Save As"));
-      gtk_widget_set_can_default (button, TRUE);
-      gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_SAVE_AS);
-      gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_SAVE_AS);
+      gtk_dialog_add_action_widget (GTK_DIALOG (dialog),
+                                    mousepad_util_image_button ("edit-delete", _("_Don't Save")),
+                                    MOUSEPAD_RESPONSE_DONT_SAVE);
+
+      /* we show different text and save button depending on the file status */
+      if (G_UNLIKELY (readonly))
+        {
+          text = _("If you don't save this read-only document as another file"
+                   ", all the changes will be lost.");
+          button = mousepad_util_image_button ("document-save-as", MOUSEPAD_LABEL_SAVE_AS);
+          button_response = MOUSEPAD_RESPONSE_SAVE_AS;
+        }
+      else
+        {
+          text = _("If you don't save the document, all the changes will be lost.");
+          button = mousepad_util_image_button ("document-save", MOUSEPAD_LABEL_SAVE);
+          button_response = MOUSEPAD_RESPONSE_SAVE;
+        }
+
+      /* set secondary text */
+      gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s", text);
     }
   else
     {
-      image = gtk_image_new_from_icon_name ("document-save", GTK_ICON_SIZE_DIALOG);
-      button = mousepad_util_image_button ("document-save", _("_Save"));
-      gtk_widget_set_can_default (button, TRUE);
-      gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_SAVE);
-      gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_SAVE);
+      button = mousepad_util_image_button ("document-save-as", MOUSEPAD_LABEL_SAVE_AS);
+      button_response = MOUSEPAD_RESPONSE_SAVE_AS;
     }
 
-  /* the content area */
-  area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
-  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 8);
-  gtk_box_pack_start (GTK_BOX (area), hbox, TRUE, TRUE, 6);
-  gtk_widget_show (hbox);
-
-  /* the dialog icon */
-  gtk_box_pack_start (GTK_BOX (hbox), image, TRUE, TRUE, 0);
-  gtk_widget_show (image);
-
-  /* the dialog message */
-  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 8);
-  gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 6);
-  gtk_widget_show (vbox);
-
-  /* primary text */
-  label = gtk_label_new (NULL);
-  gtk_label_set_markup (GTK_LABEL (label),
-                        _("<big><b>Do you want to save the changes before closing?</b></big>"));
-  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, TRUE, 0);
-  gtk_widget_show (label);
-
-  /* secondary text */
-  label = gtk_label_new (_("If you don't save the document, all the changes will be lost."));
-  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, TRUE, 0);
-  gtk_widget_show (label);
+  gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, button_response);
+  gtk_widget_set_can_default (button, TRUE);
+  gtk_dialog_set_default_response (GTK_DIALOG (dialog), button_response);
 
   /* run the dialog and wait for a response */
   response = gtk_dialog_run (GTK_DIALOG (dialog));
@@ -462,7 +456,7 @@ mousepad_dialogs_externally_modified (GtkWindow *parent,
       text_1 = _("The document has been externally modified. Do you want to continue saving?");
       text_2 = _("If you save the document, all of the external changes will be lost.");
       icon = "document-save-as";
-      label = _("Save _As");
+      label = MOUSEPAD_LABEL_SAVE_AS;
       button_response = MOUSEPAD_RESPONSE_SAVE_AS;
     }
   else
@@ -474,13 +468,13 @@ mousepad_dialogs_externally_modified (GtkWindow *parent,
         {
           text_2 = _("You have unsaved changes. If you revert the file, they will be lost.");
           icon = "document-revert";
-          label = _("_Revert");
+          label = MOUSEPAD_LABEL_REVERT;
         }
       else
         {
           text_2 = NULL;
           icon = "view-refresh";
-          label = _("Re_load");
+          label = MOUSEPAD_LABEL_RELOAD;
         }
     }
 
@@ -489,13 +483,17 @@ mousepad_dialogs_externally_modified (GtkWindow *parent,
                                                GTK_BUTTONS_NONE, "<b><big>%s</big></b>", text_1);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
-  /* add title */
+  /* setup CSD titlebar */
   gtk_window_set_title (GTK_WINDOW (dialog), _("Externally Modified"));
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
+  /* set secondary text */
   if (text_2 != NULL)
     gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s", text_2);
 
   /* add buttons */
-  gtk_dialog_add_buttons (GTK_DIALOG (dialog), _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL, NULL);
+  gtk_dialog_add_buttons (GTK_DIALOG (dialog), MOUSEPAD_LABEL_CANCEL,
+                          MOUSEPAD_RESPONSE_CANCEL, NULL);
   gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_CANCEL);
 
   button = mousepad_util_image_button (icon, label);
@@ -503,7 +501,7 @@ mousepad_dialogs_externally_modified (GtkWindow *parent,
 
   if (saving)
     {
-      button = mousepad_util_image_button ("document-save", _("_Save"));
+      button = mousepad_util_image_button ("document-save", MOUSEPAD_LABEL_SAVE);
       gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_SAVE);
     }
 
@@ -521,8 +519,7 @@ mousepad_dialogs_externally_modified (GtkWindow *parent,
 gint
 mousepad_dialogs_revert (GtkWindow *parent)
 {
-  GtkWidget *dialog;
-  GtkWidget *button;
+  GtkWidget *dialog, *button;
   gint       response;
 
   /* setup the question dialog */
@@ -531,17 +528,23 @@ mousepad_dialogs_revert (GtkWindow *parent)
                                    _("Do you want to save your changes before reloading?"));
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  gtk_window_set_title (GTK_WINDOW (dialog), _("Save Changes"));
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* set subtitle */
   gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-                                   _("If you revert the file, all unsaved changes will be lost."));
+    _("If you revert the file, all unsaved changes will be lost."));
 
   /* add buttons */
-  gtk_dialog_add_buttons (GTK_DIALOG (dialog), _("_Cancel"), MOUSEPAD_RESPONSE_CANCEL, NULL);
+  gtk_dialog_add_buttons (GTK_DIALOG (dialog), MOUSEPAD_LABEL_CANCEL,
+                          MOUSEPAD_RESPONSE_CANCEL, NULL);
+  gtk_dialog_set_default_response (GTK_DIALOG (dialog), MOUSEPAD_RESPONSE_CANCEL);
 
-  button = mousepad_util_image_button ("document-save-as", _("_Save As"));
+  button = mousepad_util_image_button ("document-save-as", MOUSEPAD_LABEL_SAVE_AS);
   gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_SAVE_AS);
 
-  button = mousepad_util_image_button ("document-revert", _("_Revert"));
+  button = mousepad_util_image_button ("document-revert", MOUSEPAD_LABEL_REVERT);
   gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, MOUSEPAD_RESPONSE_RELOAD);
 
   /* run the dialog */
@@ -574,7 +577,10 @@ mousepad_dialogs_confirm_encoding (const gchar *charset,
                                    charset, user_charset);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
-  /* set subtitle */
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
+  /* set secondary text */
   gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
                                             _("If not, the guessed encoding will be used."));
 
@@ -589,6 +595,192 @@ mousepad_dialogs_confirm_encoding (const gchar *charset,
 
 
 
+static gboolean
+mousepad_dialogs_session_restore_quit_idle (gpointer data)
+{
+  g_action_activate (data, NULL);
+
+  return FALSE;
+}
+
+
+
+static void
+mousepad_dialogs_session_restore_quit (GSimpleAction *action,
+                                       GVariant      *parameter,
+                                       GtkDialog     *dialog)
+{
+  /* disconnect this handler */
+  mousepad_disconnect_by_func (action, mousepad_dialogs_session_restore_quit, dialog);
+
+  /* cancel dialog, which results in restoring files, and activate "app.quit" again */
+  mousepad_dialogs_response_cancel (dialog);
+  g_idle_add (mousepad_dialogs_session_restore_quit_idle,
+              mousepad_util_source_autoremove (action));
+}
+
+
+
+gint
+mousepad_dialogs_session_restore (void)
+{
+  GtkApplication *application;
+  GtkWindow      *parent;
+  GtkWidget      *dialog;
+  GAction        *action;
+  gint            response;
+
+  /* get the parent window */
+  application = GTK_APPLICATION (g_application_get_default ());
+  parent = gtk_application_get_active_window (application);
+
+  /* setup the dialog */
+  dialog = gtk_message_dialog_new (parent, GTK_DIALOG_MODAL,
+                                   GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
+                                   _("It seems that the previous session did not end normally."
+                                     " Do you want to restore the available data?"));
+
+  /* destroy with parent or on "app.quit" */
+  if (parent != NULL)
+    mousepad_dialogs_destroy_with_parent (dialog, parent);
+  else
+    {
+      action = g_action_map_lookup_action (G_ACTION_MAP (application), "quit");
+      g_signal_connect_object (action, "activate",
+                               G_CALLBACK (mousepad_dialogs_session_restore_quit), dialog, 0);
+    }
+
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
+  /* set secondary text */
+  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+                                            _("If not, all this data will be lost."));
+
+  /* run the dialog */
+  response = gtk_dialog_run (GTK_DIALOG (dialog));
+
+  /* destroy the dialog */
+  gtk_widget_destroy (dialog);
+
+  return response;
+}
+
+
+
+static gboolean
+mousepad_dialogs_combo_set_active (GtkComboBox      *combo,
+                                   MousepadEncoding  encoding)
+{
+  GtkTreeModel     *model;
+  GtkTreeIter       iter;
+  MousepadEncoding  value;
+  gint              n = 0;
+
+  model = gtk_combo_box_get_model (combo);
+  do
+    {
+      if (! gtk_tree_model_iter_nth_child (model, &iter, NULL, n++))
+        return FALSE;
+
+      gtk_tree_model_get (model, &iter, 1, &value, -1);
+    }
+  while (value != encoding);
+
+  gtk_combo_box_set_active_iter (combo, &iter);
+
+  return TRUE;
+}
+
+
+
+static void
+mousepad_dialogs_open_selection_changed (GtkFileChooser *chooser,
+                                         GtkComboBox    *combo)
+{
+  MousepadEncoding  encoding = MOUSEPAD_ENCODING_NONE;
+  GtkTreeModel     *model;
+  GtkTreeIter       iter;
+  GSList           *files;
+  gchar            *label = NULL;
+  gint              row_type, row = 1;
+  gboolean          replace = FALSE;
+
+  /* do nothing if no file or several files are selected */
+  files = gtk_file_chooser_get_files (chooser);
+  if (files == NULL || g_slist_length (files) > 1)
+    {
+      g_slist_free_full (files, g_object_unref);
+      return;
+    }
+
+  /* get encoding from history, exit on failure */
+  mousepad_history_recent_get_encoding (files->data, &encoding);
+  if (encoding == MOUSEPAD_ENCODING_NONE)
+    {
+      g_slist_free_full (files, g_object_unref);
+      return;
+    }
+
+  /* get second row type */
+  model = gtk_combo_box_get_model (combo);
+  gtk_tree_model_iter_nth_child (model, &iter, NULL, row);
+  gtk_tree_model_get (model, &iter, 1, &row_type, -1);
+
+  /* shorten list */
+  if (row_type == -1)
+    {
+      /* if already in the list, set history encoding active and update label if needed */
+      if (mousepad_dialogs_combo_set_active (combo, encoding))
+        {
+          gtk_combo_box_get_active_iter (combo, &iter);
+          gtk_tree_model_get (model, &iter, 0, &label, -1);
+          if (g_strstr_len (label, -1, "(") == NULL)
+            {
+              g_free (label);
+              label = g_strdup_printf ("%s (%s)", MOUSEPAD_ENCODING_LABEL_HISTORY,
+                                       mousepad_encoding_get_charset (encoding));
+              gtk_list_store_set (GTK_LIST_STORE (model), &iter, 0, label, -1);
+            }
+
+          g_free (label);
+        }
+      /* not in the list, add it */
+      else
+        {
+          /* insert after all predefined encodings */
+          do
+            {
+              row++;
+              gtk_tree_model_iter_next (model, &iter);
+              g_free (label);
+              gtk_tree_model_get (model, &iter, 0, &label, -1);
+            }
+          while (label != NULL && g_strstr_len (label, -1, "(") != NULL
+                 && ! (replace = g_strstr_len (label, -1, MOUSEPAD_ENCODING_LABEL_HISTORY) != NULL));
+
+          g_free (label);
+          label = g_strdup_printf ("%s (%s)", MOUSEPAD_ENCODING_LABEL_HISTORY,
+                                   mousepad_encoding_get_charset (encoding));
+          if (replace)
+            gtk_list_store_set (GTK_LIST_STORE (model), &iter, 0, label, 1, encoding, -1);
+          else
+            gtk_list_store_insert_with_values (GTK_LIST_STORE (model), NULL, row,
+                                               0, label, 1, encoding, -1);
+
+          gtk_combo_box_set_active (combo, row);
+          g_free (label);
+        }
+    }
+  /* full list */
+  else
+    mousepad_dialogs_combo_set_active (combo, encoding);
+
+  g_slist_free_full (files, g_object_unref);
+}
+
+
+
 static gboolean
 mousepad_dialogs_combo_insert_separator (GtkTreeModel *model,
                                          GtkTreeIter  *iter,
@@ -607,9 +799,7 @@ mousepad_dialogs_combo_insert_separator (GtkTreeModel *model,
 static gboolean
 mousepad_dialogs_combo_popup (gpointer data)
 {
-  /* the combo box may no longer exist when we get here */
-  if (GTK_IS_COMBO_BOX (data))
-    gtk_combo_box_popup (GTK_COMBO_BOX (data));
+  gtk_combo_box_popup (data);
 
   return FALSE;
 }
@@ -644,17 +834,23 @@ mousepad_dialogs_combo_changed (GtkComboBox *combo,
   if (row_type == -2)
     {
       /* fallback to default encoding as a last resort */
-      gtk_combo_box_set_active (combo, 2);
+      mousepad_dialogs_combo_set_active (combo, mousepad_encoding_get_default ());
 
       /* "Open" dialog */
       if (gtk_file_chooser_get_action (GTK_FILE_CHOOSER (dialog))
           == GTK_FILE_CHOOSER_ACTION_OPEN)
         {
+          /* fallback to history encoding if possible */
+          mousepad_dialogs_open_selection_changed (GTK_FILE_CHOOSER (dialog), combo);
+          gtk_combo_box_get_active_iter (combo, &iter);
+          gtk_tree_model_get (model, &iter, 1, &encoding, -1);
+
           /* get a list of selected locations, must be non empty */
           if ((files = gtk_file_chooser_get_files (GTK_FILE_CHOOSER (dialog))) != NULL)
             {
               file = mousepad_file_new (GTK_TEXT_BUFFER (gtk_source_buffer_new (NULL)));
               mousepad_file_set_location (file, files->data, TRUE);
+              mousepad_file_set_encoding (file, encoding);
               g_slist_free_full (files, g_object_unref);
             }
           /* ask the user to select a file */
@@ -664,13 +860,18 @@ mousepad_dialogs_combo_changed (GtkComboBox *combo,
       /* "Save As" dialog */
       else
         {
+          /* fallback to current encoding if possible */
+          current_file = mousepad_object_get_data (dialog, "file");
+          encoding = mousepad_file_get_encoding (current_file);
+          if (encoding != MOUSEPAD_ENCODING_NONE)
+            mousepad_dialogs_combo_set_active (combo, encoding);
+
           /* try to create a temporary file to save the current buffer */
           if ((g_file = g_file_new_tmp (NULL, &iostream, &error)) != NULL)
             {
-              current_file = mousepad_object_get_data (dialog, "file");
               file = mousepad_file_new (mousepad_file_get_buffer (current_file));
               mousepad_file_set_location (file, g_file, TRUE);
-              if ((encoding = mousepad_file_get_encoding (current_file)) != MOUSEPAD_ENCODING_NONE)
+              if (encoding != MOUSEPAD_ENCODING_NONE)
                 mousepad_file_set_encoding (file, encoding);
               else
                 mousepad_file_set_encoding (file, mousepad_encoding_get_default ());
@@ -762,7 +963,6 @@ mousepad_dialogs_combo_changed (GtkComboBox *combo,
 
           /* set full list, spreading it over several columns */
           gtk_combo_box_set_model (combo, GTK_TREE_MODEL (list));
-          gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 2);
           n_rows = MOUSEPAD_N_ENCODINGS + 1;
           gtk_combo_box_set_wrap_width (combo, n_rows / 10 + (n_rows % 10 != 0));
         }
@@ -771,7 +971,6 @@ mousepad_dialogs_combo_changed (GtkComboBox *combo,
         {
           /* set shorten list */
           gtk_combo_box_set_model (combo, short_model);
-          gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 2);
           gtk_combo_box_set_wrap_width (combo, 1);
 
           /* cleanup */
@@ -779,8 +978,21 @@ mousepad_dialogs_combo_changed (GtkComboBox *combo,
           short_model = NULL;
         }
 
+      /* select a reasonable encoding */
+      mousepad_dialogs_combo_set_active (combo, mousepad_encoding_get_default ());
+      if (gtk_file_chooser_get_action (GTK_FILE_CHOOSER (dialog)) == GTK_FILE_CHOOSER_ACTION_OPEN)
+        mousepad_dialogs_open_selection_changed (GTK_FILE_CHOOSER (dialog), combo);
+      else
+        {
+          current_file = mousepad_object_get_data (dialog, "file");
+          encoding = mousepad_file_get_encoding (current_file);
+          if (encoding != MOUSEPAD_ENCODING_NONE)
+            mousepad_dialogs_combo_set_active (combo, encoding);
+        }
+
       /* reopen combo box */
-      g_idle_add_full (G_PRIORITY_LOW, mousepad_dialogs_combo_popup, combo, NULL);
+      g_idle_add_full (G_PRIORITY_LOW, mousepad_dialogs_combo_popup,
+                       mousepad_util_source_autoremove (combo), NULL);
     }
 }
 
@@ -794,7 +1006,6 @@ mousepad_dialogs_add_encoding_combo (GtkWidget *dialog)
   GtkWidget        *hbox, *widget, *combo;
   GtkListStore     *list;
   GtkCellRenderer  *cell;
-  const gchar      *charset;
   gchar            *label;
   guint             n_rows = 0, n;
   MousepadEncoding  encodings[] = { MOUSEPAD_ENCODING_UTF_8, MOUSEPAD_ENCODING_ISO_8859_15 };
@@ -817,17 +1028,17 @@ mousepad_dialogs_add_encoding_combo (GtkWidget *dialog)
 
   /* add default charset */
   default_encoding = mousepad_encoding_get_default ();
-  label = g_strdup_printf ("%s (%s)", _("Default"),
+  label = g_strdup_printf ("%s (%s)", MOUSEPAD_ENCODING_LABEL_DEFAULT,
                            mousepad_encoding_get_charset (default_encoding));
   gtk_list_store_insert_with_values (list, NULL, n_rows++, 0, label, 1, default_encoding, -1);
   g_free (label);
 
   /* add system charset if supported and different from default */
-  g_get_charset (&charset);
-  system_encoding = mousepad_encoding_find (charset);
+  system_encoding = mousepad_encoding_get_system ();
   if (system_encoding != MOUSEPAD_ENCODING_NONE && system_encoding != default_encoding)
     {
-      label = g_strdup_printf ("%s (%s)", _("System"), charset);
+      label = g_strdup_printf ("%s (%s)", MOUSEPAD_ENCODING_LABEL_SYSTEM,
+                               mousepad_encoding_get_charset (system_encoding));
       gtk_list_store_insert_with_values (list, NULL, n_rows++, 0, label, 1, system_encoding, -1);
       g_free (label);
     }
@@ -876,8 +1087,11 @@ mousepad_dialogs_add_encoding_combo (GtkWidget *dialog)
   gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), cell, TRUE);
   gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), cell, "text", 0, NULL);
 
-  /* select default encoding */
-  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 2);
+  /* select start encoding */
+  if (current_encoding != MOUSEPAD_ENCODING_NONE)
+    mousepad_dialogs_combo_set_active (GTK_COMBO_BOX (combo), current_encoding);
+  else
+    mousepad_dialogs_combo_set_active (GTK_COMBO_BOX (combo), default_encoding);
 
   return GTK_COMBO_BOX (combo);
 }
@@ -918,11 +1132,14 @@ mousepad_dialogs_save_as (GtkWindow         *parent,
 
   /* create the dialog */
   dialog = gtk_file_chooser_dialog_new (_("Save As"), parent, GTK_FILE_CHOOSER_ACTION_SAVE,
-                                        _("_Cancel"), GTK_RESPONSE_CANCEL, NULL);
+                                        MOUSEPAD_LABEL_CANCEL, GTK_RESPONSE_CANCEL, NULL);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* add button */
-  button = mousepad_util_image_button ("document-save", _("_Save"));
+  button = mousepad_util_image_button ("document-save", MOUSEPAD_LABEL_SAVE);
   gtk_widget_set_can_default (button, TRUE);
   gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_ACCEPT);
 
@@ -978,9 +1195,12 @@ mousepad_dialogs_open (GtkWindow         *parent,
 
   /* create new file chooser dialog */
   dialog = gtk_file_chooser_dialog_new (_("Open File"), parent, GTK_FILE_CHOOSER_ACTION_OPEN,
-                                        _("_Cancel"), GTK_RESPONSE_CANCEL, NULL);
+                                        MOUSEPAD_LABEL_CANCEL, GTK_RESPONSE_CANCEL, NULL);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* add button */
   button = mousepad_util_image_button ("document-open", _("_Open"));
   gtk_widget_set_can_default (button, TRUE);
@@ -997,6 +1217,10 @@ mousepad_dialogs_open (GtkWindow         *parent,
   /* encoding selector */
   combo = mousepad_dialogs_add_encoding_combo (dialog);
 
+  /* update encoding selector from history when selection changes */
+  g_signal_connect (dialog, "selection-changed",
+                    G_CALLBACK (mousepad_dialogs_open_selection_changed), combo);
+
   /* select the active document in the file chooser */
   if (file != NULL && g_file_query_exists (file, NULL))
     gtk_file_chooser_set_file (GTK_FILE_CHOOSER (dialog), file, NULL);
@@ -1030,6 +1254,9 @@ mousepad_dialogs_select_font (GtkWindow *parent)
   dialog = gtk_font_chooser_dialog_new (_("Choose Mousepad Font"), parent);
   mousepad_dialogs_destroy_with_parent (dialog, parent);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* set the current font */
   if ((font = MOUSEPAD_SETTING_GET_STRING (FONT)) != NULL)
     {
