@@ -72,6 +72,13 @@ enum
   LAST_SIGNAL
 };
 
+enum
+{
+  INIT,
+  VISIBLE,
+  HIDDEN,
+};
+
 struct _MousepadDocumentClass
 {
   GtkScrolledWindowClass __parent__;
@@ -90,9 +97,10 @@ struct _MousepadDocumentPrivate
   gchar                  *utf8_filename;
   gchar                  *utf8_basename;
 
-  /* search contexts */
+  /* search related */
   GtkSourceSearchContext *search_context, *selection_context;
   GtkSourceBuffer        *selection_buffer;
+  gint                    prev_search_state;
 };
 
 
@@ -132,7 +140,7 @@ mousepad_document_class_init (MousepadDocumentClass *klass)
 
   document_signals[ENCODING_CHANGED] =
     g_signal_new (I_("encoding-changed"), G_TYPE_FROM_CLASS (gobject_class), G_SIGNAL_RUN_LAST,
-                  0, NULL, NULL, g_cclosure_marshal_VOID__ENUM, G_TYPE_NONE, 1, G_TYPE_INT);
+                  0, NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
 
   document_signals[LANGUAGE_CHANGED] =
     g_signal_new (I_("language-changed"), G_TYPE_FROM_CLASS (gobject_class), G_SIGNAL_RUN_LAST,
@@ -152,43 +160,50 @@ mousepad_document_class_init (MousepadDocumentClass *klass)
 
 
 static void
-mousepad_document_post_init (MousepadDocument *document)
+mousepad_document_hierarchy_changed (MousepadDocument *document,
+                                     GtkWidget        *prev_window)
 {
   GtkWidget *window;
   gboolean   visible;
 
-  /* disconnect this handler */
-  mousepad_disconnect_by_func (document, mousepad_document_post_init, NULL);
+  /* disconnect from previous window signals in case of a drag and drop */
+  if (prev_window != NULL)
+    mousepad_disconnect_by_func (prev_window, mousepad_document_search_widget_visible, document);
 
   /* get the ancestor MousepadWindow */
   window = gtk_widget_get_ancestor (GTK_WIDGET (document), MOUSEPAD_TYPE_WINDOW);
 
-  /* there might not be a MousepadWindow ancestor, e.g. when the document is packed
-   * in a MousepadEncodingDialog */
-  if (window != NULL)
+  /* apart from the cases where the document is removed (perhaps temporarily during a
+   * drag and drop), there might also not be a MousepadWindow ancestor, e.g. when the
+   * document is packed in a MousepadEncodingDialog */
+  if (window == NULL)
+    return;
+
+  /* initialize autosave only for documents in a MousepadWindow, and only once */
+  if (prev_window == NULL)
+    mousepad_file_autosave_init (document->file);
+
+  /* bind some search properties to the "search-widget-visible" window property */
+  g_signal_connect_object (window, "notify::search-widget-visible",
+                           G_CALLBACK (mousepad_document_search_widget_visible),
+                           document, G_CONNECT_SWAPPED);
+
+  /* get the window property */
+  g_object_get (window, "search-widget-visible", &visible, NULL);
+
+  /* block search context handlers, so that they are unblocked below without warnings */
+  if (visible && document->priv->prev_search_state != VISIBLE)
     {
-      /* bind some search properties to the "search-widget-visible" window property */
-      g_signal_connect_object (window, "notify::search-widget-visible",
-                               G_CALLBACK (mousepad_document_search_widget_visible),
-                               document, G_CONNECT_SWAPPED);
-
-      /* get the window property */
-      g_object_get (window, "search-widget-visible", &visible, NULL);
-
-      /* block search context handlers, so that they are unblocked below without warnings */
-      if (visible)
-        {
-          g_signal_handlers_block_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
-                                           g_signal_lookup ("insert-text", GTK_TYPE_TEXT_BUFFER),
-                                           0, NULL, NULL, document->priv->search_context);
-          g_signal_handlers_block_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
-                                           g_signal_lookup ("delete-range", GTK_TYPE_TEXT_BUFFER),
-                                           0, NULL, NULL, document->priv->search_context);
-        }
-
-      /* initialize binding */
-      mousepad_document_search_widget_visible (document, NULL, MOUSEPAD_WINDOW (window));
+      g_signal_handlers_block_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
+                                       g_signal_lookup ("insert-text", GTK_TYPE_TEXT_BUFFER),
+                                       0, NULL, NULL, document->priv->search_context);
+      g_signal_handlers_block_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
+                                       g_signal_lookup ("delete-range", GTK_TYPE_TEXT_BUFFER),
+                                       0, NULL, NULL, document->priv->search_context);
     }
+
+  /* initialize binding */
+  mousepad_document_search_widget_visible (document, NULL, MOUSEPAD_WINDOW (window));
 }
 
 
@@ -200,7 +215,8 @@ mousepad_document_init (MousepadDocument *document)
   GtkSourceSearchSettings *search_settings;
 
   /* we will complete initialization when the document is anchored */
-  g_signal_connect (document, "hierarchy-changed", G_CALLBACK (mousepad_document_post_init), NULL);
+  g_signal_connect (document, "hierarchy-changed",
+                    G_CALLBACK (mousepad_document_hierarchy_changed), NULL);
 
   /* private structure */
   document->priv = mousepad_document_get_instance_private (document);
@@ -212,6 +228,7 @@ mousepad_document_init (MousepadDocument *document)
   document->priv->css_provider = gtk_css_provider_new ();
   document->priv->selection_context = NULL;
   document->priv->selection_buffer = NULL;
+  document->priv->prev_search_state = INIT;
 
   /* setup the scrolled window */
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (document),
@@ -267,8 +284,7 @@ mousepad_document_init (MousepadDocument *document)
   /* forward some document attribute signals more or less directly */
   g_signal_connect_swapped (document->buffer, "notify::cursor-position",
                             G_CALLBACK (mousepad_document_notify_cursor_position), document);
-  MOUSEPAD_SETTING_CONNECT_OBJECT (TAB_WIDTH,
-                                   G_CALLBACK (mousepad_document_notify_cursor_position),
+  MOUSEPAD_SETTING_CONNECT_OBJECT (TAB_WIDTH, mousepad_document_notify_cursor_position,
                                    document, G_CONNECT_SWAPPED);
   g_signal_connect (document->file, "encoding-changed",
                     G_CALLBACK (mousepad_document_notify_encoding), document);
@@ -280,13 +296,25 @@ mousepad_document_init (MousepadDocument *document)
 
 
 
+static gboolean
+mousepad_document_unref (gpointer data)
+{
+  g_object_unref (data);
+
+  return FALSE;
+}
+
+
+
 static void
 mousepad_document_post_finalize (GtkSourceSearchContext *search_context,
                                  GParamSpec             *pspec,
                                  gpointer                buffer)
 {
-  g_object_unref (search_context);
-  g_object_unref (buffer);
+  /* let the search context finish its work completely (no warning without that,
+   * but Valgrind sees invalid reads) */
+  g_idle_add (mousepad_document_unref, search_context);
+  g_idle_add (mousepad_document_unref, buffer);
 }
 
 
@@ -336,7 +364,7 @@ mousepad_document_finalize (GObject *object)
    * including buffers scanning, to prevent too late accesses to the buffers.
    */
   mousepad_document_finalize_search (document, document->buffer, document->priv->search_context);
-  if (GTK_SOURCE_IS_BUFFER (document->priv->selection_buffer))
+  if (document->priv->selection_buffer != NULL)
     mousepad_document_finalize_search (document, document->priv->selection_buffer,
                                        document->priv->selection_context);
 
@@ -349,7 +377,7 @@ static void
 mousepad_document_notify_cursor_position (MousepadDocument *document)
 {
   GtkTextIter iter;
-  gint        line, column, selection, tab_size;
+  gint        line, column, selection;
 
   g_return_if_fail (MOUSEPAD_IS_DOCUMENT (document));
 
@@ -360,11 +388,8 @@ mousepad_document_notify_cursor_position (MousepadDocument *document)
   /* get the current line number */
   line = gtk_text_iter_get_line (&iter) + 1;
 
-  /* get the tab size */
-  tab_size = MOUSEPAD_SETTING_GET_INT (TAB_WIDTH);
-
   /* get the column */
-  column = mousepad_util_get_real_line_offset (&iter, tab_size);
+  column = mousepad_util_get_real_line_offset (&iter);
 
   /* get length of the selection */
   selection = mousepad_view_get_selection_length (document->textview);
@@ -479,7 +504,16 @@ mousepad_document_location_changed (MousepadDocument *document,
   g_return_if_fail (file != NULL);
 
   /* convert the title into a utf-8 valid version for display */
-  utf8_filename = g_filename_to_utf8 (mousepad_util_get_path (file), -1, NULL, NULL, NULL);
+  utf8_filename = (gchar *) mousepad_util_get_path (file);
+  if (G_LIKELY (utf8_filename != NULL))
+    utf8_filename = g_filename_to_utf8 (utf8_filename, -1, NULL, NULL, NULL);
+  else
+    {
+      utf8_short_filename = g_file_get_uri (file);
+      utf8_filename = g_uri_unescape_string (utf8_short_filename, NULL);
+      g_free (utf8_short_filename);
+    }
+
   if (G_LIKELY (utf8_filename))
     {
       /* create a shorter display filename: replace $HOME with a tilde if user is not root */
@@ -576,6 +610,20 @@ mousepad_document_focus_textview (MousepadDocument *document)
 
 
 
+static void
+mousepad_document_expand_tabs_changed (MousepadDocument *document)
+{
+  gboolean expand;
+
+  expand = MOUSEPAD_SETTING_GET_BOOLEAN (EXPAND_TABS);
+
+  gtk_widget_set_hexpand (document->priv->label, expand);
+  gtk_label_set_ellipsize (GTK_LABEL (document->priv->label),
+                           expand ? PANGO_ELLIPSIZE_MIDDLE : PANGO_ELLIPSIZE_NONE);
+}
+
+
+
 GtkWidget *
 mousepad_document_get_tab_label (MousepadDocument *document)
 {
@@ -594,7 +642,9 @@ mousepad_document_get_tab_label (MousepadDocument *document)
 
   /* create the label */
   document->priv->label = gtk_label_new (mousepad_document_get_basename (document));
-  gtk_label_set_ellipsize (GTK_LABEL (document->priv->label), PANGO_ELLIPSIZE_MIDDLE);
+  mousepad_document_expand_tabs_changed (document);
+  MOUSEPAD_SETTING_CONNECT_OBJECT (EXPAND_TABS, mousepad_document_expand_tabs_changed,
+                                   document, G_CONNECT_SWAPPED);
   gtk_container_add (GTK_CONTAINER (document->priv->ebox), document->priv->label);
   gtk_widget_show (document->priv->label);
 
@@ -659,22 +709,28 @@ mousepad_document_search_completed (GObject      *object,
                                     GAsyncResult *result,
                                     gpointer      data)
 {
-  MousepadDocument        *document;
+  MousepadDocument        *document = data;
+  MousepadSearchFlags      flags;
   GtkSourceSearchContext  *search_context = GTK_SOURCE_SEARCH_CONTEXT (object);
   GtkSourceSearchSettings *search_settings;
   GtkTextBuffer           *selection_buffer;
-  MousepadSearchFlags      flags;
   GtkTextIter              iter, start, end;
+  GError                  *error = NULL;
   gchar                   *selected_text;
   const gchar             *string, *replace;
   gboolean                 found;
 
-  /* exit if the document or the view were destroyed during the search process */
-  if (! MOUSEPAD_IS_DOCUMENT (data)
-      || ! MOUSEPAD_IS_VIEW (MOUSEPAD_DOCUMENT (data)->textview))
-    return;
-  else
-    document = MOUSEPAD_DOCUMENT (data);
+  /* exit if the document was removed during the search process */
+  if (gtk_widget_get_parent (GTK_WIDGET (document)) == NULL)
+    {
+      /* let the cancellable operation reach its end before removing the last reference */
+      g_idle_add (mousepad_document_unref, document);
+
+      return;
+    }
+
+  /* remove extra reference kept at first stage */
+  g_object_unref (document);
 
   /* retrieve the first stage data */
   flags = GPOINTER_TO_INT (mousepad_object_get_data (search_context, "flags"));
@@ -691,19 +747,23 @@ mousepad_document_search_completed (GObject      *object,
 #if GTK_SOURCE_MAJOR_VERSION >= 4
   if (flags & MOUSEPAD_SEARCH_FLAGS_DIR_BACKWARD)
     found = gtk_source_search_context_backward_finish (search_context, result,
-                                                       &start, &end, NULL, NULL);
+                                                       &start, &end, NULL, &error);
   else
     found = gtk_source_search_context_forward_finish (search_context, result,
-                                                      &start, &end, NULL, NULL);
+                                                      &start, &end, NULL, &error);
 #else
   if (flags & MOUSEPAD_SEARCH_FLAGS_DIR_BACKWARD)
     found = gtk_source_search_context_backward_finish2 (search_context, result,
-                                                        &start, &end, NULL, NULL);
+                                                        &start, &end, NULL, &error);
   else
     found = gtk_source_search_context_forward_finish2 (search_context, result,
-                                                       &start, &end, NULL, NULL);
+                                                       &start, &end, NULL, &error);
 #endif
 
+  /* exit if the operation was cancelled */
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
+
   /* force the signal emission, to cover cases where Mousepad search settings change without
    * changing GtkSourceView settings (e.g. when switching between single-document mode and
    * multi-document mode) */
@@ -753,7 +813,9 @@ mousepad_document_search_completed (GObject      *object,
             }
         }
     }
-  else if (! (flags & MOUSEPAD_SEARCH_FLAGS_AREA_SELECTION))
+  /* deselect previous result when the new search fails or the search field is reset */
+  else if (! (flags & MOUSEPAD_SEARCH_FLAGS_ACTION_NONE)
+           && ! (flags & MOUSEPAD_SEARCH_FLAGS_AREA_SELECTION))
     gtk_text_buffer_place_cursor (document->buffer, &iter);
 }
 
@@ -768,6 +830,7 @@ mousepad_document_search (MousepadDocument    *document,
   GtkSourceSearchContext  *search_context;
   GtkSourceSearchSettings *search_settings, *search_settings_doc;
   GtkTextIter              iter, start, end;
+  GCancellable            *cancellable;
   gchar                   *selected_text;
   const gchar             *reference = "";
   gboolean                 has_references;
@@ -783,7 +846,7 @@ mousepad_document_search (MousepadDocument    *document,
   if (flags & MOUSEPAD_SEARCH_FLAGS_AREA_SELECTION)
     {
       /* initialize selection buffer and search context if needed */
-      if (! GTK_SOURCE_IS_BUFFER (document->priv->selection_buffer))
+      if (document->priv->selection_buffer == NULL)
         {
           document->priv->selection_buffer = gtk_source_buffer_new (NULL);
           document->priv->selection_context =
@@ -854,13 +917,18 @@ mousepad_document_search (MousepadDocument    *document,
   mousepad_object_set_data_full (search_context, "replace",
                                  g_strconcat (reference, replace, NULL), g_free);
 
+  /* keep the document alive during the search process */
+  g_object_ref (document);
+
   /* search the string */
+  cancellable = g_cancellable_new ();
   if (flags & MOUSEPAD_SEARCH_FLAGS_DIR_BACKWARD)
-    gtk_source_search_context_backward_async (search_context, &iter, NULL,
+    gtk_source_search_context_backward_async (search_context, &iter, cancellable,
                                               mousepad_document_search_completed, document);
   else
-    gtk_source_search_context_forward_async (search_context, &iter, NULL,
+    gtk_source_search_context_forward_async (search_context, &iter, cancellable,
                                              mousepad_document_search_completed, document);
+  g_object_unref (cancellable);
 }
 
 
@@ -947,8 +1015,11 @@ mousepad_document_search_widget_visible (MousepadDocument *document,
   /* get the search settings */
   search_settings = gtk_source_search_context_get_settings (document->priv->search_context);
 
-  if (visible)
+  if (visible && document->priv->prev_search_state != VISIBLE)
     {
+      /* update previous search state */
+      document->priv->prev_search_state = VISIBLE;
+
       /* unblock search context handlers */
       g_signal_handlers_unblock_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
                                          g_signal_lookup ("insert-text", GTK_TYPE_TEXT_BUFFER),
@@ -960,10 +1031,10 @@ mousepad_document_search_widget_visible (MousepadDocument *document,
       /* activate the workaround to prevent endless buffer scanning */
       mousepad_document_prevent_endless_scanning (document, visible);
       MOUSEPAD_SETTING_CONNECT_OBJECT (SEARCH_HIGHLIGHT_ALL,
-                                       G_CALLBACK (mousepad_document_prevent_endless_scanning),
+                                       mousepad_document_prevent_endless_scanning,
                                        document, G_CONNECT_SWAPPED);
       MOUSEPAD_SETTING_CONNECT_OBJECT (SEARCH_ENABLE_REGEX,
-                                       G_CALLBACK (mousepad_document_prevent_endless_scanning),
+                                       mousepad_document_prevent_endless_scanning,
                                        document, G_CONNECT_SWAPPED);
 
       /* bind "highlight" and "regex-enabled" search settings to Mousepad settings */
@@ -972,8 +1043,11 @@ mousepad_document_search_widget_visible (MousepadDocument *document,
       MOUSEPAD_SETTING_BIND (SEARCH_ENABLE_REGEX, search_settings,
                              "regex-enabled", G_SETTINGS_BIND_GET);
     }
-  else
+  else if (! visible && document->priv->prev_search_state != HIDDEN)
     {
+      /* update previous search state */
+      document->priv->prev_search_state = HIDDEN;
+
       /* block search context handlers */
       g_signal_handlers_block_matched (document->buffer, G_SIGNAL_MATCH_DATA | G_SIGNAL_MATCH_ID,
                                        g_signal_lookup ("insert-text", GTK_TYPE_TEXT_BUFFER),
