@@ -190,6 +190,18 @@ mousepad_encoding_get_default (void)
 
 
 
+MousepadEncoding
+mousepad_encoding_get_system (void)
+{
+  const gchar *charset;
+
+  g_get_charset (&charset);
+
+  return mousepad_encoding_find (charset);
+}
+
+
+
 MousepadEncoding
 mousepad_encoding_read_bom (const gchar *contents,
                             gsize        length,
@@ -261,7 +273,7 @@ mousepad_encoding_read_bom (const gchar *contents,
 
 void
 mousepad_encoding_write_bom (MousepadEncoding  *encoding,
-                             gint              *length,
+                             gsize             *length,
                              gchar            **contents)
 {
   guchar bom[4];
