@@ -129,7 +129,7 @@ mousepad_print_finalize (GObject *object)
   MousepadPrint *print = MOUSEPAD_PRINT (object);
 
   /* cleanup */
-  if (GTK_IS_PAGE_SETUP (print->page_setup))
+  if (print->page_setup != NULL)
     g_object_unref (print->page_setup);
 
   g_object_unref (print->compositor);
@@ -144,19 +144,16 @@ mousepad_print_settings_load (GtkPrintOperation *operation)
 {
   MousepadPrint         *print = MOUSEPAD_PRINT (operation);
   GKeyFile              *keyfile;
-  gchar                 *filename;
   GtkPrintSettings      *settings = NULL;
-  gchar                **keys;
-  gint                   i;
-  gchar                 *key;
-  gchar                 *value;
-  gchar                 *body_font = NULL;
-  gchar                 *header_font = NULL;
-  gchar                 *line_numbers_font = NULL;
   GtkPageSetup          *page_setup;
   GtkPaperSize          *paper_size;
   PangoContext          *context;
   PangoFontDescription  *font_desc;
+  gchar                **keys;
+  gchar                 *filename, *key, *value, *body_font = NULL, *header_font = NULL,
+                        *line_numbers_font = NULL;
+  gint                   i;
+  gdouble                margin;
 
   g_return_if_fail (MOUSEPAD_IS_DOCUMENT (print->document));
   g_return_if_fail (GTK_IS_WIDGET (print->document->textview));
@@ -211,24 +208,32 @@ mousepad_print_settings_load (GtkPrintOperation *operation)
       /* apply the settings */
       gtk_print_operation_set_print_settings (operation, settings);
 
-      if (gtk_print_settings_get_bool (settings, "page-setup-saved") == TRUE)
+      if (gtk_print_settings_get_bool (settings, "page-setup-saved"))
         {
           /* create new page setup */
           page_setup = gtk_page_setup_new ();
 
           /* set orientation */
-          gtk_page_setup_set_orientation (page_setup, gtk_print_settings_get_orientation (settings));
+          gtk_page_setup_set_orientation (page_setup,
+            gtk_print_settings_get_orientation (settings));
 
           /* restore margins */
-          gtk_page_setup_set_top_margin (page_setup, gtk_print_settings_get_double (settings, "top-margin"), GTK_UNIT_MM);
-          gtk_page_setup_set_bottom_margin (page_setup, gtk_print_settings_get_double (settings, "bottom-margin"), GTK_UNIT_MM);
-          gtk_page_setup_set_right_margin (page_setup, gtk_print_settings_get_double (settings, "right-margin"), GTK_UNIT_MM);
-          gtk_page_setup_set_left_margin (page_setup, gtk_print_settings_get_double (settings, "left-margin"), GTK_UNIT_MM);
+          margin = gtk_print_settings_get_double (settings, "top-margin");
+          gtk_page_setup_set_top_margin (page_setup, margin, GTK_UNIT_MM);
+          margin = gtk_print_settings_get_double (settings, "bottom-margin");
+          gtk_page_setup_set_bottom_margin (page_setup, margin, GTK_UNIT_MM);
+          margin = gtk_print_settings_get_double (settings, "right-margin");
+          gtk_page_setup_set_right_margin (page_setup, margin, GTK_UNIT_MM);
+          margin = gtk_print_settings_get_double (settings, "left-margin");
+          gtk_page_setup_set_left_margin (page_setup, margin, GTK_UNIT_MM);
 
           /* set paper size */
           paper_size = gtk_print_settings_get_paper_size (settings);
           if (G_LIKELY (paper_size))
-            gtk_page_setup_set_paper_size (page_setup, paper_size);
+            {
+              gtk_page_setup_set_paper_size (page_setup, paper_size);
+              gtk_paper_size_free (paper_size);
+            }
 
           /* set the default page setup */
           gtk_print_operation_set_default_page_setup (operation, page_setup);
@@ -244,7 +249,8 @@ mousepad_print_settings_load (GtkPrintOperation *operation)
                     "print-line-numbers",
                     gtk_print_settings_get_int (settings, "line-numbers-increment"),
                     "wrap-mode",
-                    gtk_print_settings_get_bool (settings, "text-wrapping") ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE,
+                    gtk_print_settings_get_bool (settings, "text-wrapping")
+                      ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE,
                     "highlight-syntax",
                     gtk_print_settings_get_bool (settings, "highlight-syntax"),
                     NULL);
@@ -287,7 +293,8 @@ mousepad_print_settings_load (GtkPrintOperation *operation)
 
     /* setup line number printing */
     if (print->print_line_numbers)
-      gtk_source_print_compositor_set_print_line_numbers (print->compositor, print->line_number_increment);
+      gtk_source_print_compositor_set_print_line_numbers (print->compositor,
+                                                          print->line_number_increment);
     else
       gtk_source_print_compositor_set_print_line_numbers (print->compositor, 0);
 
@@ -323,91 +330,94 @@ mousepad_print_settings_save (GtkPrintOperation *operation)
 {
   MousepadPrint    *print = MOUSEPAD_PRINT (operation);
   GKeyFile         *keyfile;
-  gchar            *filename;
   GtkPrintSettings *settings;
   GtkPageSetup     *page_setup;
   GtkPaperSize     *paper_size;
+  gchar            *filename, *str;
+  gboolean          value;
 
   /* get the save location */
   filename = mousepad_util_get_save_location (MOUSEPAD_RC_RELPATH, TRUE);
+  if (G_UNLIKELY (filename == NULL))
+    return;
+
+  /* get the print settings */
+  settings = gtk_print_operation_get_print_settings (operation);
+  if (G_UNLIKELY (settings == NULL))
+    {
+      g_free (filename);
+      return;
+    }
+
+  /* get the page setup */
+  page_setup = gtk_print_operation_get_default_page_setup (operation);
+
+  /* restore the page setup */
+  if (G_LIKELY (page_setup != NULL))
+    {
+      /* the the settings page orienation */
+      gtk_print_settings_set_orientation (settings,
+        gtk_page_setup_get_orientation (page_setup));
+
+      /* save margins */
+      gtk_print_settings_set_double (settings, "top-margin",
+        gtk_page_setup_get_top_margin (page_setup, GTK_UNIT_MM));
+      gtk_print_settings_set_double (settings, "bottom-margin",
+        gtk_page_setup_get_bottom_margin (page_setup, GTK_UNIT_MM));
+      gtk_print_settings_set_double (settings, "right-margin",
+        gtk_page_setup_get_right_margin (page_setup, GTK_UNIT_MM));
+      gtk_print_settings_set_double (settings, "left-margin",
+        gtk_page_setup_get_left_margin (page_setup, GTK_UNIT_MM));
+
+      /* get the paper size */
+      paper_size = gtk_page_setup_get_paper_size (page_setup);
+
+      /* set settings page size */
+      if (G_LIKELY (paper_size))
+        gtk_print_settings_set_paper_size (settings, paper_size);
+    }
+
+  /* a bool we use for loading */
+  gtk_print_settings_set_bool (settings, "page-setup-saved", page_setup != NULL);
+
+  /* set print settings */
+  value = gtk_source_print_compositor_get_print_header (print->compositor);
+  gtk_print_settings_set_bool (settings, "print-header", value);
+
+  gtk_print_settings_set_bool (settings, "print-line-numbers", print->print_line_numbers);
+
+  gtk_print_settings_set_int (settings, "line-numbers-increment", print->line_number_increment);
+
+  value = gtk_source_print_compositor_get_wrap_mode (print->compositor) == GTK_WRAP_NONE ? FALSE
+                                                                                         : TRUE;
+  gtk_print_settings_set_bool (settings, "text-wrapping", value);
+
+  value = gtk_source_print_compositor_get_highlight_syntax (print->compositor);
+  gtk_print_settings_set_bool (settings, "highlight-syntax", value);
+
+  str = gtk_source_print_compositor_get_body_font_name (print->compositor);
+  gtk_print_settings_set (settings, "body-font-name", str);
+  g_free (str);
+
+  str = gtk_source_print_compositor_get_header_font_name (print->compositor);
+  gtk_print_settings_set (settings, "header-font-name", str);
+  g_free (str);
+
+  str = gtk_source_print_compositor_get_line_numbers_font_name (print->compositor);
+  gtk_print_settings_set (settings, "line-numbers-font-name", str);
+  g_free (str);
 
   /* create a new keyfile */
   keyfile = g_key_file_new ();
 
   /* load the existing settings */
-  if (G_LIKELY (g_key_file_load_from_file (keyfile, filename, G_KEY_FILE_NONE, NULL)))
-    {
-      /* get the print settings */
-      settings = gtk_print_operation_get_print_settings (operation);
+  g_key_file_load_from_file (keyfile, filename, G_KEY_FILE_NONE, NULL);
 
-      if (G_LIKELY (settings != NULL))
-        {
-          /* get the page setup */
-          page_setup = gtk_print_operation_get_default_page_setup (operation);
+  /* store all the print settings */
+  gtk_print_settings_foreach (settings, mousepad_print_settings_save_foreach, keyfile);
 
-          /* restore the page setup */
-          if (G_LIKELY (page_setup != NULL))
-            {
-              /* the the settings page orienation */
-              gtk_print_settings_set_orientation (settings, gtk_page_setup_get_orientation (page_setup));
-
-              /* save margins */
-              gtk_print_settings_set_double (settings, "top-margin", gtk_page_setup_get_top_margin (page_setup, GTK_UNIT_MM));
-              gtk_print_settings_set_double (settings, "bottom-margin", gtk_page_setup_get_bottom_margin (page_setup, GTK_UNIT_MM));
-              gtk_print_settings_set_double (settings, "right-margin", gtk_page_setup_get_right_margin (page_setup, GTK_UNIT_MM));
-              gtk_print_settings_set_double (settings, "left-margin", gtk_page_setup_get_left_margin (page_setup, GTK_UNIT_MM));
-
-              /* get the paper size */
-              paper_size = gtk_page_setup_get_paper_size (page_setup);
-
-              /* set settings page size */
-              if (G_LIKELY (paper_size))
-                gtk_print_settings_set_paper_size (settings, paper_size);
-            }
-
-          /* a bool we use for loading */
-          gtk_print_settings_set_bool (settings, "page-setup-saved", page_setup != NULL);
-
-          /* set print settings */
-          gtk_print_settings_set_bool (settings,
-                                       "print-header",
-                                       gtk_source_print_compositor_get_print_header (print->compositor));
-
-          gtk_print_settings_set_bool (settings,
-                                       "print-line-numbers",
-                                       print->print_line_numbers);
-
-          gtk_print_settings_set_int (settings,
-                                      "line-numbers-increment",
-                                      print->line_number_increment);
-
-          gtk_print_settings_set_bool (settings,
-                                       "text-wrapping",
-                                       gtk_source_print_compositor_get_wrap_mode (print->compositor) == GTK_WRAP_NONE ? FALSE : TRUE);
-
-          gtk_print_settings_set_bool (settings,
-                                       "highlight-syntax",
-                                       gtk_source_print_compositor_get_highlight_syntax (print->compositor));
-
-          gtk_print_settings_set (settings,
-                                  "body-font-name",
-                                  gtk_source_print_compositor_get_body_font_name (print->compositor));
-
-          gtk_print_settings_set (settings,
-                                  "header-font-name",
-                                  gtk_source_print_compositor_get_header_font_name (print->compositor));
-
-          gtk_print_settings_set (settings,
-                                  "line-numbers-font-name",
-                                  gtk_source_print_compositor_get_line_numbers_font_name (print->compositor));
-
-          /* store all the print settings */
-          gtk_print_settings_foreach (settings, mousepad_print_settings_save_foreach, keyfile);
-
-          /* save the contents */
-          mousepad_util_save_key_file (keyfile, filename);
-        }
-    }
+  /* save the contents */
+  mousepad_util_save_key_file (keyfile, filename);
 
   /* cleanup */
   g_key_file_free (keyfile);
@@ -426,7 +436,7 @@ mousepad_print_begin_print (GtkPrintOperation *operation,
   const gchar      *file_name;
 
   /* apply the saved page setup, if any */
-  if (GTK_IS_PAGE_SETUP (print->page_setup))
+  if (print->page_setup != NULL)
     gtk_print_operation_set_default_page_setup (operation, print->page_setup);
 
   /* print header */
@@ -496,7 +506,7 @@ mousepad_print_page_setup_dialog (GtkWidget         *button,
    * this dialog and begin_print(). So let's save it here as a MousepadPrint attribute
    * and apply it later in begin_print().
    */
-  if (GTK_IS_PAGE_SETUP (print->page_setup))
+  if (print->page_setup != NULL)
     g_object_unref (print->page_setup);
 
   print->page_setup = gtk_print_run_page_setup_dialog (GTK_WINDOW (toplevel), page_setup, settings);
@@ -518,12 +528,14 @@ mousepad_print_button_toggled (GtkWidget     *button,
     print->print_line_numbers = active;
     gtk_widget_set_sensitive (print->widget_line_numbers_hbox, active);
     if (active)
-      gtk_source_print_compositor_set_print_line_numbers (print->compositor, print->line_number_increment);
+      gtk_source_print_compositor_set_print_line_numbers (print->compositor,
+                                                          print->line_number_increment);
     else
       gtk_source_print_compositor_set_print_line_numbers (print->compositor, 0);
   }
   else if (button == print->widget_text_wrapping)
-    gtk_source_print_compositor_set_wrap_mode (print->compositor, active ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
+    gtk_source_print_compositor_set_wrap_mode (print->compositor,
+                                               active ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
   else if (button == print->widget_syntax_highlighting)
     gtk_source_print_compositor_set_highlight_syntax (print->compositor, active);
 }
@@ -590,12 +602,9 @@ static GtkWidget *
 mousepad_print_create_custom_widget (GtkPrintOperation *operation)
 {
   MousepadPrint *print = MOUSEPAD_PRINT (operation);
-  GtkWidget     *button;
-  GtkWidget     *vbox, *vbox2;
-  GtkWidget     *frame;
-  GtkWidget     *label;
-  GtkWidget     *grid;
+  GtkWidget     *button, *vbox, *vbox2, *frame, *label, *grid;
   GtkAdjustment *adjustment;
+  gchar         *str;
 
   vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 8);
@@ -610,7 +619,8 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_frame_set_label_widget (GTK_FRAME (frame), label);
   gtk_widget_show (label);
 
-  button = mousepad_util_image_button ("document-properties", _("_Adjust page size and orientation"));
+  button = mousepad_util_image_button ("document-properties",
+                                       _("_Adjust page size and orientation"));
   g_signal_connect (button, "clicked", G_CALLBACK (mousepad_print_page_setup_dialog), operation);
   gtk_widget_set_halign (button, GTK_ALIGN_START);
   gtk_widget_set_margin_start (button, 12);
@@ -639,14 +649,16 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_container_add (GTK_CONTAINER (frame), vbox2);
   gtk_widget_show (vbox2);
 
-  button = print->widget_page_headers = gtk_check_button_new_with_mnemonic (_("Print page _headers"));
+  button = print->widget_page_headers =
+    gtk_check_button_new_with_mnemonic (_("Print page _headers"));
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
                                 gtk_source_print_compositor_get_print_header (print->compositor));
   g_signal_connect (button, "toggled", G_CALLBACK (mousepad_print_button_toggled), print);
   gtk_box_pack_start (GTK_BOX (vbox2), button, FALSE, FALSE, 0);
   gtk_widget_show (button);
 
-  button = print->widget_line_numbers = gtk_check_button_new_with_mnemonic (_("Print _line numbers"));
+  button = print->widget_line_numbers =
+    gtk_check_button_new_with_mnemonic (_("Print _line numbers"));
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
                                 print->print_line_numbers);
   g_signal_connect (button, "toggled", G_CALLBACK (mousepad_print_button_toggled), print);
@@ -683,7 +695,8 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
                       print->widget_line_numbers_spin, FALSE, TRUE, 0);
   gtk_widget_show (print->widget_line_numbers_spin);
 
-  button = print->widget_text_wrapping = gtk_check_button_new_with_mnemonic (_("Enable text _wrapping"));
+  button = print->widget_text_wrapping =
+    gtk_check_button_new_with_mnemonic (_("Enable text _wrapping"));
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
                                 gtk_source_print_compositor_get_wrap_mode (print->compositor)
                                   == GTK_WRAP_NONE ? FALSE : TRUE);
@@ -691,7 +704,8 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_box_pack_start (GTK_BOX (vbox2), button, FALSE, FALSE, 0);
   gtk_widget_show (button);
 
-  button = print->widget_syntax_highlighting = gtk_check_button_new_with_mnemonic (_("Enable _syntax highlighting"));
+  button = print->widget_syntax_highlighting =
+    gtk_check_button_new_with_mnemonic (_("Enable _syntax highlighting"));
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
                                 gtk_source_print_compositor_get_highlight_syntax (print->compositor));
   g_signal_connect (button, "toggled", G_CALLBACK (mousepad_print_button_toggled), print);
@@ -725,8 +739,9 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_grid_attach (GTK_GRID (grid), label, 0, 0, 1, 1);
   gtk_widget_show (label);
 
-  print->widget_header_font = gtk_font_button_new_with_font (
-                                gtk_source_print_compositor_get_header_font_name (print->compositor));
+  str = gtk_source_print_compositor_get_header_font_name (print->compositor);
+  print->widget_header_font = gtk_font_button_new_with_font (str);
+  g_free (str);
   gtk_grid_attach (GTK_GRID (grid), print->widget_header_font, 1, 0, 1, 1);
   g_signal_connect (print->widget_header_font, "font-set",
                     G_CALLBACK (mousepad_print_button_font_set), print);
@@ -738,8 +753,9 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_grid_attach (GTK_GRID (grid), label, 0, 1, 1, 1);
   gtk_widget_show (label);
 
-  print->widget_body_font = gtk_font_button_new_with_font (
-                              gtk_source_print_compositor_get_body_font_name (print->compositor));
+  str = gtk_source_print_compositor_get_body_font_name (print->compositor);
+  print->widget_body_font = gtk_font_button_new_with_font (str);
+  g_free (str);
   gtk_grid_attach (GTK_GRID (grid), print->widget_body_font, 1, 1, 1, 1);
   g_signal_connect (print->widget_body_font, "font-set",
                     G_CALLBACK (mousepad_print_button_font_set), print);
@@ -751,7 +767,9 @@ mousepad_print_create_custom_widget (GtkPrintOperation *operation)
   gtk_grid_attach (GTK_GRID (grid), label, 0, 2, 1, 1);
   gtk_widget_show (label);
 
-  print->widget_line_numbers_font = gtk_font_button_new_with_font (gtk_source_print_compositor_get_line_numbers_font_name (print->compositor));
+  str = gtk_source_print_compositor_get_line_numbers_font_name (print->compositor);
+  print->widget_line_numbers_font = gtk_font_button_new_with_font (str);
+  g_free (str);
   gtk_grid_attach (GTK_GRID (grid), print->widget_line_numbers_font, 1, 2, 1, 1);
   g_signal_connect (print->widget_line_numbers_font, "font-set",
                     G_CALLBACK (mousepad_print_button_font_set), print);
