@@ -192,6 +192,9 @@ mousepad_replace_dialog_post_init (MousepadReplaceDialog *dialog)
   /* disconnect this handler */
   mousepad_disconnect_by_func (dialog, mousepad_replace_dialog_post_init, NULL);
 
+  /* setup CSD titlebar */
+  mousepad_util_set_titlebar (GTK_WINDOW (dialog));
+
   /* get the transient parent window and the application */
   window = gtk_window_get_transient_for (GTK_WINDOW (dialog));
   application = gtk_window_get_application (window);
@@ -464,14 +467,15 @@ mousepad_replace_dialog_reset_display (MousepadReplaceDialog *dialog)
 {
   const gchar *string;
 
-  /* reset entry color and occurrences label */
-  mousepad_util_entry_error (dialog->search_entry, FALSE);
+  /* reset occurrences label */
   gtk_label_set_text (GTK_LABEL (dialog->hits_label), NULL);
 
-  /* start the spinner */
+  /* start the spinner or reset entry color */
   string = gtk_entry_get_text (GTK_ENTRY (dialog->search_entry));
   if (string != NULL && *string != '\0')
     gtk_spinner_start (GTK_SPINNER (dialog->spinner));
+  else
+    mousepad_util_entry_error (dialog->search_entry, FALSE);
 }
 
 
@@ -493,7 +497,7 @@ mousepad_replace_dialog_response (GtkWidget *widget,
     }
 
   /* search direction */
-  search_direction = MOUSEPAD_SETTING_GET_INT (SEARCH_DIRECTION);
+  search_direction = MOUSEPAD_SETTING_GET_UINT (SEARCH_DIRECTION);
   if ((search_direction == DIRECTION_UP && response_id != MOUSEPAD_RESPONSE_REVERSE_FIND)
       || (search_direction != DIRECTION_UP && response_id == MOUSEPAD_RESPONSE_REVERSE_FIND))
     flags = MOUSEPAD_SEARCH_FLAGS_DIR_BACKWARD;
@@ -503,7 +507,7 @@ mousepad_replace_dialog_response (GtkWidget *widget,
   /* search area */
   if (MOUSEPAD_SETTING_GET_BOOLEAN (SEARCH_REPLACE_ALL))
     {
-      replace_all_location = MOUSEPAD_SETTING_GET_INT (SEARCH_REPLACE_ALL_LOCATION);
+      replace_all_location = MOUSEPAD_SETTING_GET_UINT (SEARCH_REPLACE_ALL_LOCATION);
       flags |= MOUSEPAD_SEARCH_FLAGS_ENTIRE_AREA;
       if (replace_all_location == IN_ALL_DOCUMENTS)
         flags |= MOUSEPAD_SEARCH_FLAGS_AREA_ALL_DOCUMENTS;
@@ -525,10 +529,13 @@ mousepad_replace_dialog_response (GtkWidget *widget,
     }
   else if (response_id == MOUSEPAD_RESPONSE_ENTRY_CHANGED)
     {
-      /* select the first match */
-      flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_SELECT;
+      /* select the first match if incremental search is enabled */
+      if (MOUSEPAD_SETTING_GET_BOOLEAN (SEARCH_INCREMENTAL))
+        flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_SELECT;
+      else
+        flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_NONE;
 
-      /* start at the 'beginning' of the selection (find as you type) */
+      /* start at the 'beginning' of the selection */
       if (flags & MOUSEPAD_SEARCH_FLAGS_DIR_BACKWARD)
         flags |= MOUSEPAD_SEARCH_FLAGS_ITER_SEL_END;
       else
@@ -582,7 +589,7 @@ mousepad_replace_dialog_search_completed (MousepadReplaceDialog *dialog,
   /* ... or if irrelevant settings for it are in use here, without stopping the spinner
    * in this case (we are in multi-document mode and this is a partial result) */
   else if (MOUSEPAD_SETTING_GET_BOOLEAN (SEARCH_REPLACE_ALL)
-           && MOUSEPAD_SETTING_GET_INT (SEARCH_REPLACE_ALL_LOCATION) != IN_DOCUMENT
+           && MOUSEPAD_SETTING_GET_UINT (SEARCH_REPLACE_ALL_LOCATION) != IN_DOCUMENT
            && ! (flags & (MOUSEPAD_SEARCH_FLAGS_AREA_SELECTION
                           | MOUSEPAD_SEARCH_FLAGS_AREA_ALL_DOCUMENTS)))
     return;
@@ -705,7 +712,7 @@ mousepad_replace_dialog_page_switched (MousepadReplaceDialog *dialog,
                                        GtkTextBuffer         *new_buffer)
 {
   /* disconnect from old buffer signals */
-  if (GTK_IS_TEXT_BUFFER (old_buffer))
+  if (old_buffer != NULL)
     mousepad_disconnect_by_func (old_buffer, mousepad_replace_dialog_reset_display, dialog);
 
   /* connect to new buffer signals to update display widgets on change */
