@@ -335,14 +335,15 @@ mousepad_search_bar_reset_display (MousepadSearchBar *bar)
 {
   const gchar *string;
 
-  /* reset entry color and occurrences label */
-  mousepad_util_entry_error (bar->entry, FALSE);
+  /* reset occurrences label */
   gtk_label_set_text (GTK_LABEL (bar->hits_label), NULL);
 
-  /* start the spinner */
+  /* start the spinner or reset entry color */
   string = gtk_entry_get_text (GTK_ENTRY (bar->entry));
   if (string != NULL && *string != '\0')
     gtk_spinner_start (GTK_SPINNER (bar->spinner));
+  else
+    mousepad_util_entry_error (bar->entry, FALSE);
 }
 
 
@@ -354,8 +355,11 @@ mousepad_search_bar_find_string (MousepadSearchBar   *bar,
   const gchar *string;
 
   /* always true when using the search bar */
-  flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_SELECT
-           | MOUSEPAD_SEARCH_FLAGS_WRAP_AROUND;
+  flags |= MOUSEPAD_SEARCH_FLAGS_WRAP_AROUND;
+
+  /* always select unless it is a silent search */
+  if (! (flags & MOUSEPAD_SEARCH_FLAGS_ACTION_NONE))
+    flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_SELECT;
 
   /* get the entry string */
   string = gtk_entry_get_text (GTK_ENTRY (bar->entry));
@@ -442,6 +446,9 @@ mousepad_search_bar_entry_changed (MousepadSearchBar *bar)
   flags = MOUSEPAD_SEARCH_FLAGS_ITER_SEL_START
           | MOUSEPAD_SEARCH_FLAGS_DIR_FORWARD;
 
+  if (! MOUSEPAD_SETTING_GET_BOOLEAN (SEARCH_INCREMENTAL))
+    flags |= MOUSEPAD_SEARCH_FLAGS_ACTION_NONE;
+
   /* find */
   mousepad_search_bar_find_string (bar, flags);
 }
@@ -505,7 +512,7 @@ mousepad_search_bar_page_switched (MousepadSearchBar *bar,
   g_return_if_fail (MOUSEPAD_IS_SEARCH_BAR (bar));
 
   /* disconnect from old buffer signals */
-  if (GTK_IS_TEXT_BUFFER (old_buffer))
+  if (old_buffer != NULL)
     mousepad_disconnect_by_func (old_buffer, mousepad_search_bar_reset_display, bar);
 
   /* connect to new buffer signals to update display widgets on change */
@@ -516,9 +523,9 @@ mousepad_search_bar_page_switched (MousepadSearchBar *bar,
                            G_CALLBACK (mousepad_search_bar_reset_display),
                            bar, G_CONNECT_SWAPPED);
 
-  /* run a search */
+  /* run a silent search */
   if (search)
-    mousepad_search_bar_entry_changed (bar);
+    mousepad_search_bar_find_string (bar, MOUSEPAD_SEARCH_FLAGS_ACTION_NONE);
 }
 
 
