@@ -16,8 +16,7 @@
 
 #include <mousepad/mousepad-private.h>
 #include <mousepad/mousepad-util.h>
-
-#include <xfconf/xfconf.h>
+#include <mousepad/mousepad-settings.h>
 
 #ifdef HAVE_MATH_H
 #include <math.h>
@@ -25,7 +24,6 @@
 
 
 
-#define DEFAULT_FONT "Monospace 10"
 #define FONT_FAMILY  "font-family"
 #define FONT_STYLE   "font-style"
 #define FONT_VARIANT "font-variant"
@@ -251,107 +249,6 @@ mousepad_util_key_name (const gchar *name)
 
 
 
-gchar *
-mousepad_util_utf8_strcapital (const gchar *str)
-{
-  gunichar     c;
-  const gchar *p;
-  gchar       *buf;
-  GString     *result;
-  gboolean     upper = TRUE;
-
-  g_return_val_if_fail (g_utf8_validate (str, -1, NULL), NULL);
-
-  /* create a new string */
-  result = g_string_sized_new (strlen (str));
-
-  /* walk though the string */
-  for (p = str; *p != '\0'; p = g_utf8_next_char (p))
-    {
-      /* get the unicode char */
-      c = g_utf8_get_char (p);
-
-      /* only change the case of alpha chars */
-      if (g_unichar_isalpha (c))
-        {
-          /* check case */
-          if (upper ? g_unichar_isupper (c) : g_unichar_islower (c))
-            {
-              /* currect case is already correct */
-              g_string_append_unichar (result, c);
-            }
-          else
-            {
-              /* convert the case of the char and append it */
-              buf = upper ? g_utf8_strup (p, 1) : g_utf8_strdown (p, 1);
-              g_string_append (result, buf);
-              g_free (buf);
-            }
-
-          /* next char must be lowercase */
-          upper = FALSE;
-        }
-      else
-        {
-          /* append the char */
-          g_string_append_unichar (result, c);
-
-          /* next alpha char uppercase after a space */
-          upper = g_unichar_isspace (c);
-        }
-    }
-
-  /* return the result */
-  return g_string_free (result, FALSE);
-}
-
-
-
-gchar *
-mousepad_util_utf8_stropposite (const gchar *str)
-{
-  gunichar     c;
-  const gchar *p;
-  gchar       *buf;
-  GString     *result;
-
-  g_return_val_if_fail (g_utf8_validate (str, -1, NULL), NULL);
-
-  /* create a new string */
-  result = g_string_sized_new (strlen (str));
-
-  /* walk though the string */
-  for (p = str; *p != '\0'; p = g_utf8_next_char (p))
-    {
-      /* get the unicode char */
-      c = g_utf8_get_char (p);
-
-      /* only change the case of alpha chars */
-      if (g_unichar_isalpha (c))
-        {
-          /* get the opposite case of the char */
-          if (g_unichar_isupper (c))
-            buf = g_utf8_strdown (p, 1);
-          else
-            buf = g_utf8_strup (p, 1);
-
-          /* append to the buffer */
-          g_string_append (result, buf);
-          g_free (buf);
-        }
-      else
-        {
-          /* append the char */
-          g_string_append_unichar (result, c);
-        }
-    }
-
-  /* return the result */
-  return g_string_free (result, FALSE);
-}
-
-
-
 gchar *
 mousepad_util_escape_underscores (const gchar *str)
 {
@@ -530,16 +427,115 @@ mousepad_util_dialog_update_header (GtkDialog   *dialog,
 
 
 
-gint
-mousepad_util_get_real_line_offset (const GtkTextIter *iter,
-                                    gint               tab_size)
+static void
+mousepad_util_decoration_layout_changed (GObject    *settings,
+                                         GParamSpec *pspec,
+                                         gpointer    bar)
 {
-  gint        offset = 0;
+  gchar *layout, *str;
+
+  g_object_get (settings, "gtk-decoration-layout", &layout, NULL);
+
+  /* add the icon on the far left of the header bar if not already present */
+  if (g_strstr_len (layout, -1, "icon") == NULL)
+    {
+      str = g_strstr_len (layout, -1, ":") == NULL ? ":" : ",";
+      str = g_strconcat ("icon", str, layout, NULL);
+      g_free (layout);
+      layout = str;
+    }
+
+  gtk_header_bar_set_decoration_layout (bar, layout);
+
+  g_free (layout);
+}
+
+
+
+void
+mousepad_util_set_titlebar (GtkWindow *window)
+{
+  static GtkSettings *settings = NULL;
+
+  GtkWidget       *bar;
+  GtkStyleContext *context;
+  GtkCssProvider  *provider;
+  const gchar     *title;
+  gboolean         show_close = TRUE;
+
+  /* set a default window title if needed */
+  title = gtk_window_get_title (window);
+  if (title == NULL || *title == '\0')
+    gtk_window_set_title (window, g_get_application_name ());
+
+  /* exit if the user doesn't want client-side decorations, ensuring we have at least
+   * one close button or similar in the title bar for all dialogs */
+  if (! MOUSEPAD_SETTING_GET_BOOLEAN (CLIENT_SIDE_DECORATIONS))
+    {
+      if (! GTK_IS_HEADER_BAR (gtk_window_get_titlebar (window)))
+        gtk_window_set_titlebar (window, NULL);
+
+      return;
+    }
+
+  /* use existing header bar or create a new one (including when the title bar widget
+   * is not a GtkHeaderBar) */
+  bar = gtk_window_get_titlebar (window);
+  if (! GTK_IS_HEADER_BAR (bar))
+    {
+      bar = gtk_header_bar_new ();
+      gtk_widget_show (bar);
+    }
+  /* do not change "show-close-button" if there is already a header bar, this should be
+   * handled internally by one means or another */
+  else
+    show_close = gtk_header_bar_get_show_close_button (GTK_HEADER_BAR (bar));
+
+  /* set properties */
+  gtk_header_bar_set_title (GTK_HEADER_BAR (bar), gtk_window_get_title (window));
+  gtk_header_bar_set_has_subtitle (GTK_HEADER_BAR (bar), FALSE);
+  gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (bar), show_close);
+
+  /* stay in sync with the window manager for the decoration layout */
+  if (settings == NULL)
+    {
+      settings = gtk_settings_get_default ();
+      if (settings != NULL)
+        {
+          mousepad_util_decoration_layout_changed (G_OBJECT (settings), NULL, bar);
+          g_signal_connect_object (settings, "notify::gtk-decoration-layout",
+                                   G_CALLBACK (mousepad_util_decoration_layout_changed), bar, 0);
+        }
+      else
+        gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (bar), "icon,menu:minimize,maximize,close");
+    }
+  else
+    mousepad_util_decoration_layout_changed (G_OBJECT (settings), NULL, bar);
+
+  /* make the header bar slim */
+  context = gtk_widget_get_style_context (bar);
+  provider = gtk_css_provider_new ();
+  gtk_css_provider_load_from_data (provider, "headerbar { min-height: 0px; }", -1, NULL);
+  gtk_style_context_add_provider (context, GTK_STYLE_PROVIDER (provider),
+                                  GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
+  g_object_unref (provider);
+
+  gtk_window_set_titlebar (window, bar);
+}
+
+
+
+gint
+mousepad_util_get_real_line_offset (const GtkTextIter *iter)
+{
+  gint        tab_size, offset = 0;
   GtkTextIter needle = *iter;
 
   /* move the needle to the start of the line */
   gtk_text_iter_set_line_offset (&needle, 0);
 
+  tab_size = MOUSEPAD_SETTING_GET_UINT (TAB_WIDTH);
+
   /* forward the needle until we hit the iter */
   while (!gtk_text_iter_equal (&needle, iter))
     {
@@ -560,16 +556,17 @@ mousepad_util_get_real_line_offset (const GtkTextIter *iter,
 
 void
 mousepad_util_set_real_line_offset (GtkTextIter *iter,
-                                    gint         tab_size,
                                     gint         column,
                                     gboolean     from_end)
 {
-  gint        char_offset = 0, column_offset = 0;
+  gint        tab_size, char_offset = 0, column_offset = 0;
   GtkTextIter needle = *iter;
 
   /* move the needle to the start of the line */
   gtk_text_iter_set_line_offset (&needle, 0);
 
+  tab_size = MOUSEPAD_SETTING_GET_UINT (TAB_WIDTH);
+
   /* forward the needle until we reach column or the end of the line */
   while (!gtk_text_iter_ends_line (&needle) && column_offset < column)
     {
@@ -603,28 +600,36 @@ mousepad_util_set_real_line_offset (GtkTextIter *iter,
 
 
 
-gboolean
-mousepad_util_forward_iter_to_text (GtkTextIter       *iter,
-                                    const GtkTextIter *limit)
+void
+mousepad_util_place_cursor (GtkTextBuffer *buffer,
+                            gint           line,
+                            gint           column)
 {
-  gunichar c;
+  GtkTextIter pos;
+  gboolean    from_end = FALSE;
 
-  do
+  /* for negative line number count from bottom */
+  if (line < 0)
+    line = MAX (gtk_text_buffer_get_line_count (buffer) + line, 0);
+
+  /* set the position iter line */
+  gtk_text_buffer_get_iter_at_line (buffer, &pos, line);
+
+  /* for negative column number count from end of the line */
+  if (column < 0)
     {
-      /* get the iter character */
-      c = gtk_text_iter_get_char (iter);
+      if (!gtk_text_iter_ends_line (&pos))
+        gtk_text_iter_forward_to_line_end (&pos);
 
-      /* break if the character is not a space */
-      if (!g_unichar_isspace (c) || c == '\n' || c == '\r')
-        break;
-
-      /* break when we reached the limit iter */
-      if (limit && gtk_text_iter_equal (iter, limit))
-        return FALSE;
+      column = MAX (mousepad_util_get_real_line_offset (&pos) + column + 1, 0);
+      from_end = TRUE;
     }
-  while (gtk_text_iter_forward_char (iter));
 
-  return TRUE;
+  /* set the position iter offset */
+  mousepad_util_set_real_line_offset (&pos, column, from_end);
+
+  /* set the cursor at the position iter */
+  gtk_text_buffer_place_cursor (buffer, &pos);
 }
 
 
@@ -641,7 +646,7 @@ mousepad_util_get_save_location (const gchar *relpath,
   filename = g_build_filename (g_get_user_config_dir (), relpath, NULL);
 
   /* test if the file exists */
-  if (G_UNLIKELY (g_file_test (filename, G_FILE_TEST_EXISTS) == FALSE))
+  if (G_UNLIKELY (! g_file_test (filename, G_FILE_TEST_EXISTS)))
     {
       if (create_parents)
         {
@@ -690,7 +695,7 @@ mousepad_util_save_key_file (GKeyFile    *keyfile,
   if (G_LIKELY (error == NULL))
     {
       /* write the contents to the file */
-      if (G_UNLIKELY (g_file_set_contents (filename, contents, length, &error) == FALSE))
+      if (G_UNLIKELY (! g_file_set_contents (filename, contents, length, &error)))
         goto print_error;
     }
   else
@@ -781,9 +786,9 @@ mousepad_util_style_schemes_name_compare (gconstpointer a,
 {
   const gchar *name_a, *name_b;
 
-  if (G_UNLIKELY (!GTK_SOURCE_IS_STYLE_SCHEME (a)))
+  if (G_UNLIKELY (a == NULL))
     return - (a != b);
-  if (G_UNLIKELY (!GTK_SOURCE_IS_STYLE_SCHEME (b)))
+  if (G_UNLIKELY (b == NULL))
     return a != b;
 
   name_a = gtk_source_style_scheme_get_name (GTK_SOURCE_STYLE_SCHEME (a));
@@ -804,6 +809,9 @@ mousepad_util_get_style_schemes (void)
   schemes = gtk_source_style_scheme_manager_get_scheme_ids (
               gtk_source_style_scheme_manager_get_default ());
 
+  if (G_UNLIKELY (schemes == NULL))
+    return NULL;
+
   while (*schemes)
     {
       scheme = gtk_source_style_scheme_manager_get_scheme (
@@ -843,7 +851,7 @@ mousepad_util_get_sorted_language_sections (void)
   while (*languages)
     {
       language = gtk_source_language_manager_get_language (manager, *languages);
-      if (G_LIKELY (GTK_SOURCE_IS_LANGUAGE (language)))
+      if (G_LIKELY (language != NULL))
         {
           /* ignore hidden languages */
           if (gtk_source_language_get_hidden (language))
@@ -874,9 +882,9 @@ mousepad_util_languages_name_compare (gconstpointer a,
 {
   const gchar *name_a, *name_b;
 
-  if (G_UNLIKELY (!GTK_SOURCE_IS_LANGUAGE (a)))
+  if (G_UNLIKELY (a == NULL))
     return - (a != b);
-  if (G_UNLIKELY (!GTK_SOURCE_IS_LANGUAGE (b)))
+  if (G_UNLIKELY (b == NULL))
     return a != b;
 
   name_a = gtk_source_language_get_name (GTK_SOURCE_LANGUAGE (a));
@@ -900,10 +908,13 @@ mousepad_util_get_sorted_languages_for_section (const gchar *section)
   manager = gtk_source_language_manager_get_default ();
   languages = gtk_source_language_manager_get_language_ids (manager);
 
+  if (G_UNLIKELY (languages == NULL))
+    return NULL;
+
   while (*languages)
     {
       language = gtk_source_language_manager_get_language (manager, *languages);
-      if (G_LIKELY (GTK_SOURCE_IS_LANGUAGE (language)))
+      if (G_LIKELY (language != NULL))
         {
           /* ignore hidden languages */
           if (gtk_source_language_get_hidden (language))
@@ -924,22 +935,6 @@ mousepad_util_get_sorted_languages_for_section (const gchar *section)
 
 
 
-gchar *
-mousepad_util_get_default_font (void)
-{
-  gchar *font;
-
-  /* get the default font from xfconf */
-  font = xfconf_channel_get_string (xfconf_channel_get ("xsettings"),
-                                    "/Gtk/MonospaceFontName", NULL);
-  if (G_UNLIKELY (font == NULL))
-    font = g_strdup (DEFAULT_FONT);
-
-  return font;
-}
-
-
-
 /*
  * Copied from Gedit 3.38.0 and slightly modified:
  * https://gitlab.gnome.org/GNOME/gedit/-/blob/21fac3f0c87db0db104d7af7eaeb6f63d8216a14/gedit/gedit-pango.c#L37
@@ -1167,3 +1162,30 @@ G_GNUC_END_IGNORE_DEPRECATIONS
 
 #endif
 }
+
+
+
+static void
+mousepad_util_source_remove_all (gpointer  data,
+                                 GObject  *object)
+{
+  while (g_source_remove_by_user_data (object));
+}
+
+
+
+/* an often sufficient way to automate memory management of sources, without having
+ * to store a source id and use an ad hoc handler */
+gpointer
+mousepad_util_source_autoremove (gpointer object)
+{
+  g_return_val_if_fail (G_IS_OBJECT (object), object);
+
+  if (! mousepad_object_get_data (object, "source-autoremove"))
+    {
+      g_object_weak_ref (object, mousepad_util_source_remove_all, NULL);
+      mousepad_object_set_data (object, "source-autoremove", GINT_TO_POINTER (TRUE));
+    }
+
+  return object;
+}
